# -*- coding: utf-8 -*-
import logging

import numpy as np
import pylab as plt
from scipy.constants import c, k, h
from ..mathmatics import trapz_inte_edge
# from . import sub
# from ..spectra.lineshape import DLV
from ..spectra.source_function import Bv_T, PopuSource_AB
from ..spectra.abscoeff import basic
from ..rtc import SOSC, FOSC
from ..spectra.lineshape import DopplerWind
from typhon.physics.em import planck, rayleighjeans


logger = logging.getLogger(__name__)


def calcu_grid(radius, alt_ref, angle=False, speed=None):
    """ Calculation grid for given altitude and radius

    Parameters:
        radius: Radius of planet/moon/object [meters]
        alt_ref: Reference altitude grid (1D grid) [kilometers]
        angle: Will return the angle grid of the atmospheric layers
        speed: Using wind-velocities (1D grid) [meters/second]

    Returns:
        Grid point weights, central altitudes, tangent point, [angle at grid points]
    """
    mu_delta = radius/100. # *1000.  # m
    mu_point = np.r_[np.arange(0, radius, mu_delta),
                     alt_ref*1e3 + radius]  # [m]
    mu_tangent = (mu_point - radius) * 1e-3

    PSC = ((alt_ref*1.e3 + radius)**2 -
           mu_point.reshape(mu_point.size, 1)**2)**0.5
    PSC2 = np.zeros((mu_point.size, alt_ref.size-1))  # alt_center

    for xx in range(1, alt_ref.size):  # alt_center
        PSC2[:, -xx] = PSC[:, -xx] - PSC[:, -xx-1]
        
    PSC_sin = mu_point.reshape(mu_point.size, 1)/(alt_ref*1.e3 + radius)
    deg = np.arcsin(PSC_sin)
    PSC_sin[deg != deg] = 0
    deg[deg != deg] = np.pi / 2.
    mu_weight = np.zeros((mu_point.size, alt_ref.size))
    
    for i in range(alt_ref.size):
        mu_weight[:, i] = trapz_inte_edge(PSC_sin[:, i], deg[:, i])
    
    if angle is True:
        return mu_weight, PSC2, mu_tangent, PSC_sin
    elif speed is not None:
        velocity = speed * np.cos(deg)
        velocity[deg == np.pi / 2.] = 0
        return mu_weight, PSC2, mu_tangent, velocity
    else:
        #tete
        return mu_weight, PSC2, mu_tangent


class MolecularConsts:
    """
    From a specific file, read followings;
    Transition frequency
    number of energy levels
    energy levels
    degeneracies
    Einstein's A coeff
    transition tag (upper and lower level for each transition)
    """
    def __init__(self, molecules_state_consts):
        self.filename = molecules_state_consts
        fn = open(molecules_state_consts + 'oH2O16.lev_7levels', 'rb')
        self.fr = fn.readlines()
        fn.close()
        self.ni = int(self.fr[3])  # NUMBER OF ENERGY LEVELS
        self.e_cm1 = np.array([float(self.fr[xx].split()[1])
                               for xx in range(5, 5 + self.ni)])
        self.weighti = np.array([float(self.fr[xx].split()[2])
                                 for xx in range(5, 5 + self.ni)])

        Ai = np.zeros((self.ni, self.ni))
        Freqi = np.zeros((self.ni, self.ni))
        E_Ki = np.zeros((self.ni, self.ni))
        B_place = np.ones((self.ni, self.ni))*-1.
        Tran_tag = []
        Ai_tag = []
        Nt = int(self.fr[3+self.ni+3])  # int(9)
        for xx in range(8+self.ni, 8+self.ni+Nt):  # Nt Number of transition ,24,twolevel ->16):
            """up: temp[1]-1, low: temp[2]-1"""
            temp = np.array(self.fr[xx].split()).astype(np.float64)
            Tran_tag.append(temp[:3]-1)  # for 1... -> 0....
            Ai[int(temp[1]-1), int(temp[2])-1] = temp[3]
            Ai_tag.append(temp[3])
            # Freqi[int(temp[1]-1), int(temp[2])-1] = temp[4]
            _deltaEcm1 = (self.e_cm1[int(temp[1])-1] -
                          self.e_cm1[int(temp[2])-1])
            Freqi[int(temp[1]-1), int(temp[2])-1] = (_deltaEcm1*c *
                                                     100.*1.e-9)  # GHz
            E_Ki[int(temp[1]-1), int(temp[2])-1] = temp[5]
            B_place[int(temp[1]-1), int(temp[2])-1] = int(temp[0]-1)  # induced
            B_place[int(temp[2])-1, int(temp[1]-1)] = int(temp[0]-1)  # abs.ed.
        Tran_tag = np.array(Tran_tag)*1
        Ai_tag = np.array(Ai_tag)*1.
        Nt = len(Tran_tag)  # int(9)
        self.ai = Ai
        self.freqi = Freqi
        self.tran_tag = Tran_tag
        self.b_place = B_place
        self.ai_tag = Ai_tag
        self.nt = Nt
        Freq_array = np.array([])
        for xx in range(Nt):
            _up = int(Tran_tag[xx][1])
            _low = int(Tran_tag[xx][2])
            Freq_array = np.hstack((Freq_array, Freqi[_up, _low]))
        self.freq_array = Freq_array
        Bul = self.ai*c**2/(2*h*(self.freqi*1.e9)**3)
        Blu = Bul*self.weighti.reshape((self.ni, 1))/self.weighti
        Bul[Bul != Bul] = 0  # [i,j]: i->j
        Blu[Blu != Blu] = 0  # [i,j]: j->i
        self.blu = Blu
        self.bul = Bul

    def ratematrix(self):
        """
        Make a matrix of radiation rate
            lower
          |     |  |n1|
        u |     |  |n2|
          |     |  |n3|
        *diagonal is sum of loss rate
        """
        Bul = self.ai*c**2/(2*h*(self.freqi*1.e9)**3)
        Blu = Bul*self.weighti.reshape((self.ni, 1))/self.weighti
        Bul[Bul != Bul] = 0  # [i,j]: i->j
        Blu[Blu != Blu] = 0  # [i,j]: j->i
        # RaRaB_induced = -Bul + Bul.T
        # RaRaB_absorption = Blu-Blu.T
        RaRaB_induced = Bul.T  # [i,j] i<-j
        RaRaB_absorption = Blu  # [i,j] i<-j
        RaRaA = self.ai.T
        # Diagonal
        RaRaAd = np.zeros((self.ni, self.ni))  #
        RaRaBd = np.zeros((self.ni, self.ni))  #
        for xx in range(self.ni):
            cond = self.tran_tag[:, 1] == xx
            RaRaAd[xx, xx] += self.ai_tag[cond].sum()*-1.
            RaRaBd[xx, xx] += (Bul[xx]+Blu.T[xx]).sum()*-1.
        return RaRaA, RaRaAd, RaRaBd, RaRaB_induced, RaRaB_absorption

    def collision(self, Mole, Temp, ETS=False):
        """
        make self collisional rate matrix
        
        input parameter;
        number of Molecules [molecules/m3] collisional parameter
        Temperature [K]
        rate constant is read from filename
        
        If ETS is True, 
        Population of escape to space approximation is plotted
        
        """
        RateConst = np.loadtxt(self.filename+'H2O'+'_col.txt')*1.e-6
        colt = np.array([100,  200,  300,  400,  500,  600,  700,  800])*1.
        coltt = np.around(np.arange(0., 400., 0.01), 2)  # We calculate 0.1 K
        col_tk = []
        for xx in range(self.nt):
            col_tk.append(np.interp(coltt, colt, RateConst[xx]))
        col_tk = np.array(col_tk)
        "set col rate for each atmospheric Temp."
        col_ul, col_lu = [], []
        for xx in range(self.nt):
            _up = int(self.tran_tag[xx][1])
            _low = int(self.tran_tag[xx][2])
            col_ul.append(np.array([col_tk[xx][np.argwhere(coltt ==
                                               round(xxx, 2))][0][0]
                                   for xxx in np.around(Temp, 2)]))
            col_lu.append((col_ul[xx]*self.weighti[_up]/self.weighti[_low]) *
                          np.exp(-h*self.freqi[_up, _low] *
                                 1.e9/k/Temp))
        rate_ul, rate_lu = Mole*col_ul, Mole*col_lu  # molecules/s
        self.Cul = rate_ul
        J_ETS, J_LTE = [], []
        for xx in range(self.nt):
            _up = int(self.tran_tag[xx][1])
            _low = int(self.tran_tag[xx][2])
            J_ETS.append(2.*h*(self.freqi[_up, _low]*1.e9)**3 /
                         c**2 /
                         (((rate_lu[xx])/(rate_ul[xx]) *
                           (rate_ul[xx]/self.ai[_up, _low]) /
                           (1.+rate_ul[xx]/self.ai[_up, _low]))**-1 *
                          self.weighti[_up]/self.weighti[_low] - 1.)
                         )  # See (5,2)(3.35)
            J_LTE.append(2.*h*(self.freqi[_up, _low]*1.e9)**3 /
                         c**2 /
                         ((rate_ul[xx])/(rate_lu[xx]) *
                          self.weighti[_up]/self.weighti[_low] - 1.)
                         )  # See (3.35)
            # P.plot(h*(Freqi[_up, _low]*1.e9/(k*np.log(2.*h*
            #             (Freqi[_up, _low]*1.e9)**3/c**2/J_LTE[xx]+1))))

            # P.plot(h*(Freqi[_up, _low]*1.e9/(k*np.log(2.*h*
            #            (Freqi[_up, _low]*1.e9)**3/c**2/J_ETS[xx] +1)))
            #       , label = str(Freqi[_up, _low]))
        CoRa_block = []  # Collisional Rate matrix
        for hoge in range(Temp.size):
            Cul, Clu = rate_ul.T[hoge], rate_lu.T[hoge]
            CoRa = np.zeros((self.ni, self.ni))
            for xx in range(self.nt):
                _up = int(self.tran_tag[xx][1])
                _low = int(self.tran_tag[xx][2])
                # CoRa[_up, _low] += -Cul[xx]+Clu[xx]
                # CoRa[_low, _up] += Cul[xx]-Clu[xx]
                CoRa[_up, _low] += Clu[xx]
                CoRa[_low, _up] += Cul[xx]
            # Diagonal
            for xx in range(self.ni):
                cond = self.tran_tag[:, 1] == xx
                CoRa[xx, xx] -= Cul[cond].sum()
                cond = self.tran_tag[:, 2] == xx
                CoRa[xx, xx] -= Clu[cond].sum()
            CoRa_block.append(CoRa*1.)
        CoRa_block = np.array(CoRa_block)*1.
        if ETS is True:
            J_ETS = np.array(J_ETS)
            J_ETS_plot = h*(self.freq_array.reshape(self.nt, 1)*1.e9 /
                            (k*np.log(2.*h *
                                      (self.freq_array.reshape(self.nt, 1) *
                                       1.e9)**3/c**2/J_ETS + 1.)))
            plt.figure()
            [plt.plot(J_ETS_plot[xx],
                      label=str(round(self.freq_array[xx],
                                      3))) for xx in range(self.nt)]
            plt.plot(Temp, 'k', label='T$_k$')
            plt.legend(loc=0)
            plt.grid()
        # ======================================================================
            return CoRa_block, J_ETS_plot

        return CoRa_block

    def population(self, Mole, Temp):
        """Calculate population for each level at given temperature"""
        RoTemp = np.reshape(Temp*1., (Temp.size, 1))
        Qr = self.weighti*np.exp(-(self.e_cm1*100*c*h)/(k*RoTemp))
        # RoPr = Qr/Ntotal  # This is for all transitions
        RoPr = Qr/(Qr.sum(axis=1).reshape(RoTemp.size, 1))  # only given trans.
        linet = []
        for xx in range(self.nt):
            gdu, gdl = self.weighti[self.tran_tag[xx][1:].astype(int)]
            _up = int(self.tran_tag[xx][1])
            _low = int(self.tran_tag[xx][2])
            Aei = self.ai[_up, _low]
            line_const = (c*10**2)**2*Aei*(gdu/gdl)*1.e-6*1.e14 /\
                         (8.*np.pi*(self.freq_array[xx]*1.e9)**2)
            # Hz->MHz,cm^2 ->nm^2
            # W = C.h*C.c*E_cm1[_low]*100.  # energy level above ground state
            "This is the function of calculating H2O intensity"
            line = (1.-np.exp(-h*(self.freq_array[xx]*1.e9) /
                              k/RoTemp))*line_const
            linet.append(line[:, 0]*RoPr[:, _low])  # line intensity non-LTE
        Ni_LTE = Mole.reshape((Mole.size, 1))*RoPr  # *0.75  # orth para ratio
        Ite_pop = [[Ni_LTE[i].reshape((self.ni, 1))] for i in range(Mole.size)]
        return Ite_pop

#import numba


"""Population Calculation
"""

""" Takayoshi Yamada:  Work in progress below.  Change sparingly.
"""
def Calc(Ite_pop, Abs_ite,
         PSC2, Mu_tangent, mu_weight,
         Alt_ref, Temp,
         Fre_range_i, Freq_array, F_vl_i,
         B_place,
         Nt, Ni, Aul, Bul, Blu,
         RaRaB_absorption, RaRaB_induced, RaRaA, RaRaAd, CoRa_block,
         Tran_tag,
         wind_v=False,
         update_population=True,
         out_put_spectra=False,
         continuum_surface_temperature_unit='Planck',
         back_ground_radiation='CMB',
         iteration='MUGA'):
#    if iteration is 'MUGA1SC':
#        from ..rtc import SOSCdamy as SOSC
#        print('Hey using FOSC in SOSC')
#        iteration = 'MUGA'
    ji_in_all = np.zeros((Nt,
                          Alt_ref.size,
                          Mu_tangent.size,
                          Fre_range_i[0].size))
    ji_out_all = ji_in_all*0.
    lambda_approx_in = np.zeros((Nt,
                                 Alt_ref.size,
                                 Mu_tangent.size,
                                 Fre_range_i[0].size))
    lambda_approx_out = np.zeros((Nt,
                                  Alt_ref.size,
                                  Mu_tangent.size,
                                  Fre_range_i[0].size))
    # for IteNum in range(0, 100):
    max_true_error = np.ones(Alt_ref.size)*0.
    new_pop = np.array(Ite_pop)*0.+0.
    for ii in range(Alt_ref.size)[::-1]:  # spatial points
        # print(ii)
        for xx in range(Nt):  # transitions
            # Line_Shape = F_vl_i[xx]
            Para=[Freq_array[xx]*1.e9, 18.0153]
            Abs_coeff = Abs_ite[xx]
            Fre_range = Fre_range_i[xx]
            up_tag = int(Tran_tag[xx][1])
            low_tag = int(Tran_tag[xx][2])
            # Up_pop = Ite_pop[ii, 0, up_tag, 0]
            # Low_pop = Ite_pop[ii, 0, low_tag, 0]
            if ii == Alt_ref.size-1:  # upper Boundary(u>0) incoming
                if back_ground_radiation == 'CMB':
                    B_v_cosmic = planck(Fre_range, 2.725)
                else:
                    B_v_cosmic = 0
                ji_in_all[xx, ii, :, :] = B_v_cosmic  # Ladi comparison
            elif ii == Alt_ref.size-1-1:
                tdu2 = Abs_coeff[ii+1] * \
                DopplerWind(Temp[ii+1], Fre_range_i[xx], Para, wind_v[:,ii+1], shift_direction='Red')#F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * \
                DopplerWind(Temp[ii], Fre_range_i[xx], Para, wind_v[:,ii], shift_direction='Red')#F_vl_i[xx][ii]
#                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
#                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                tdu = (0.5*np.abs(tdu1+tdu2) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :],lambda_approx_in[xx, ii, :, :]\
                         = FOSC(tdu, Sd1, Sd, Idu)
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
    # test = trapz_inte_edge(PSC_sin[:, -1], deg[:, -1])
    # PSC_cos = np.arccos(PSC/(Alt_ref*1.e3+Radius))
    # PSC_cos1 = PSC/(Alt_ref*1.e3+Radius)
    # PSC_sin = (1-PSC_cos**2)**0.5
    # PSC_cos = np.cos( np.arccos(PSC_cos1)*1.)  # degree (pi)
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
            elif ii == 0:  # SOSC is not available (FOSC elif ii>=0)
                tdu2 = Abs_coeff[ii+1] * F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * F_vl_i[xx][ii]
                tdu2 = Abs_coeff[ii+1] * \
                DopplerWind(Temp[ii+1], Fre_range_i[xx], Para, wind_v[:,ii+1], shift_direction='Red')#F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * \
                DopplerWind(Temp[ii], Fre_range_i[xx], Para, wind_v[:,ii], shift_direction='Red')#F_vl_i[xx][ii]
#                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
#                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                tdu = (0.5*np.abs(tdu1+tdu2) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd3 = PopuSource_AB(Ite_pop[ii-1, 0, low_tag, 0],
                                    Ite_pop[ii-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
#                tdu3 = Abs_coeff[ii-1] * F_vl_i[xx][ii-1]
                tdu3 = Abs_coeff[ii-1] * \
                DopplerWind(Temp[ii-1], Fre_range_i[xx], Para, wind_v[:,ii-1], shift_direction='Red')#F_vl_i[xx][ii]
#                tdb = (0.5*np.abs(tdu1+tdu3).reshape((1, Fre_range.size)) *
#                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                tdb = (0.5*np.abs(tdu1+tdu3) *
                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :],lambda_approx_in[xx, ii, :, :]\
                = FOSC(tdu, Sd1, Sd, Idu)
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
            else:  # SOSC
                tdu2 = Abs_coeff[ii+1] * F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * F_vl_i[xx][ii]
                tdu2 = Abs_coeff[ii+1] * \
                DopplerWind(Temp[ii+1], Fre_range_i[xx], Para, wind_v[:,ii+1], shift_direction='Red')#F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * \
                DopplerWind(Temp[ii], Fre_range_i[xx], Para, wind_v[:,ii], shift_direction='Red')#F_vl_i[xx][ii]
#                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
#                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                tdu = (0.5*np.abs(tdu1+tdu2) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd3 = PopuSource_AB(Ite_pop[ii-1, 0, low_tag, 0],
                                    Ite_pop[ii-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                tdu3 = Abs_coeff[ii-1] * F_vl_i[xx][ii-1]
                tdu3 = Abs_coeff[ii-1] * \
                DopplerWind(Temp[ii-1], Fre_range_i[xx], Para, wind_v[:,ii-1], shift_direction='Red')#F_vl_i[xx][ii]
#                tdb = (0.5*np.abs(tdu1+tdu3).reshape((1, Fre_range.size)) *
#                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                tdb = (0.5*np.abs(tdu1+tdu3) *
                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :], lambda_approx_in[xx, ii, :, :]\
                         = SOSC(tdu, tdb,
                                Sd1, Sd, Sd3,
                                Idu)
                ji_in_all[xx, ii, (Mu_tangent == Alt_ref[ii]), :],\
                lambda_approx_in[xx, ii, (Mu_tangent == Alt_ref[ii]), :]\
                          = FOSC(tdu[Mu_tangent == Alt_ref[ii]],
                                    Sd1,
                                    Sd,
                                    Idu[Mu_tangent == Alt_ref[ii]])
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
    for i in range(Alt_ref.size):  # spatial points
        #  print(i)
        B_int = B_place*0.  # intensity for each transitoin
        B_int_lamda = B_place*0.  # intensity for each transitoin for lambda ope
        A_int_lamda = B_place*0.  # approximated lambda operate
        SF = np.ones(Nt)
        for xx in range(Nt):  # transitions
            Para=[Freq_array[xx]*1.e9, 18.0153]
            Fre_range = Fre_range_i[xx]
            # Line_Shape = F_vl_i[xx]
            Abs_coeff = Abs_ite[xx]
            Fre_range = Fre_range_i[xx]
            up_tag = int(Tran_tag[xx][1])
            low_tag = int(Tran_tag[xx][2])
            # Up_pop = Ite_pop[i, 0, up_tag, 0]
            # Low_pop = Ite_pop[i, 0, low_tag, 0]
            if i == 0:  # Lower Boundary(u>0) outgoing
                if continuum_surface_temperature_unit is 'Planck':
#                    ji_out_all[xx, i, :, :] = planck(Fre_range, Temp[i])
                    ji_out_all[xx, i, :, :] = Bv_T(Fre_range, Temp[i])
                elif continuum_surface_temperature_unit is 'RJ':
                    ji_out_all[xx, i, :, :] = rayleighjeans(Fre_range, Temp[i])
                elif continuum_surface_temperature_unit is 'RJ_obs':
                    _t_phys = h * Freq_array[xx] * 1.e9 / \
                              (k * np.log(h * Freq_array[xx] * 1.e9
                                          /k/Temp[i] + 1))
                    ji_out_all[xx, i, :, :] = planck(Fre_range, _t_phys)
                else:
                    ji_out_all[xx, i, :, :] = Bv_T(Fre_range, Temp[i])
                ji_out_all[:, i, (Mu_tangent > Alt_ref[i]), :] = 0
            elif i < Alt_ref.size-1:  # SOSC
                tl_1 = Abs_coeff[i] * F_vl_i[xx][i]
                # GS-iteration
                if iteration is 'MUGA':
                    tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                                 new_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                elif iteration is 'LI' or iteration is 'MALI':
                    tl_2 = basic(Ite_pop[i-1, 0, low_tag, 0],
                                 Ite_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                # tdl2 = calc_abscoeff(xx, IteNum+1, i-1) * F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2).reshape((1, Fre_range.size)) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                tl_3 = Abs_coeff[i+1] * F_vl_i[xx][i+1]
                tl3 = (0.5*np.abs(tl_1+tl_3).reshape((1, Fre_range.size)) *
                       PSC2[:, i].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """===Doppler=========="""
                tl_1 = Abs_coeff[i] * \
                DopplerWind(Temp[i], Fre_range_i[xx], Para, wind_v[:,i], shift_direction='Blue')
                if iteration is 'MUGA':
                    tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                                 new_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*\
                           DopplerWind(Temp[i-1], Fre_range_i[xx], Para, wind_v[:,i-1], 
                                       shift_direction='Blue')  # F_vl_i[xx][i-1]
                elif iteration is 'LI' or iteration is 'MALI':
                    tl_2 = basic(Ite_pop[i-1, 0, low_tag, 0],
                                 Ite_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*\
                           DopplerWind(Temp[i-1], Fre_range_i[xx], Para, wind_v[:,i-1], 
                                       shift_direction='Blue')  # F_vl_i[xx][i-1]
                tl_3 = Abs_coeff[i+1] * \
                DopplerWind(Temp[i+1], Fre_range_i[xx], Para, wind_v[:,i+1], shift_direction='Red')
                tl1 = (0.5*np.abs(tl_1+tl_2) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                tl3 = (0.5*np.abs(tl_1+tl_3) *
                       PSC2[:, i].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """======================"""
                Sl2 = PopuSource_AB(Ite_pop[i, 0, low_tag, 0],
                                    Ite_pop[i, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                SF[xx] = Sl2
                if iteration is 'MUGA':
                    Sl1 = PopuSource_AB(new_pop[i-1, 0, low_tag, 0],
                                        new_pop[i-1, 0, up_tag, 0],
                                        Aul[up_tag, low_tag],
                                        Bul[up_tag, low_tag],
                                        Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                elif iteration is 'LI' or iteration is 'MALI':
                    Sl1 = PopuSource_AB(Ite_pop[i-1, 0, low_tag, 0],
                                        Ite_pop[i-1, 0, up_tag, 0],
                                        Aul[up_tag, low_tag],
                                        Bul[up_tag, low_tag],
                                        Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Sl3 = PopuSource_AB(Ite_pop[i+1, 0, low_tag, 0],
                                    Ite_pop[i+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])
                Il = ji_out_all[xx, i-1, :, :]*1.  #
                # ji_out_all[xx, i, :, :] = FOSC(tl1, Sl1, Sl2, Il)  # (12.113)
                ji_out_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                          = SOSC(tl1, tl3,
                                 Sl1, Sl2, Sl3,
                                 Il)
                ji_out_all[xx, i, Mu_tangent > Alt_ref[i], :]=0
                lambda_approx_out[xx, i, Mu_tangent > Alt_ref[i], :] = 0
                ji_out_all[xx, i, Mu_tangent == Alt_ref[i], :]\
                = ji_in_all[xx, i, Mu_tangent == Alt_ref[i], :]
                lambda_approx_out[xx, i, Mu_tangent == Alt_ref[i], :]\
                = lambda_approx_in[xx, i, Mu_tangent == Alt_ref[i], :]
                I3 = ji_in_all[xx, i+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                         = SOSC(tl3, tl1,
                                Sl3, Sl2, Sl1,
                                I3)
                ji_in_all[xx, i, (Mu_tangent == Alt_ref[i]), :],\
                lambda_approx_out[xx, i, (Mu_tangent == Alt_ref[i]), :]\
                             = FOSC(tl3[Mu_tangent == Alt_ref[i]],
                                    Sl3,
                                    Sl2,
                                    I3[Mu_tangent == Alt_ref[i]])
                ji_in_all[xx, i, ((Mu_tangent > Alt_ref[i])), :] = 0
                lambda_approx_out[xx, i, ((Mu_tangent > Alt_ref[i])), :] = 0
#                ji_in_all[xx, i, ((Mu_tangent > Alt_ref[i-1])), :] = 0
                if np.argwhere(Mu_tangent == Alt_ref[i]).size!=1:
                    logger.error(' '.join((i, xx, "error!")))

            elif i == Alt_ref.size-1:  # FOSC
                tl_1 = Abs_coeff[i] * F_vl_i[xx][i]
                # GS-iteration
                if iteration is 'MUGA':
                    tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                                 new_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                elif iteration is 'MALI' or iteration is 'LI':
                    tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                                 new_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                else:
                    raise ValueError(f'Invalid iteration method {iteration}')
                # tdl2 = calc_abscoeff(xx, IteNum+1, i-1) * F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2).reshape((1, Fre_range.size)) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """===Doppler=========="""
                tl_1 = (Abs_coeff[i] *
                        DopplerWind(Temp[i], Fre_range_i[xx], Para,
                                    wind_v[:, i], shift_direction='Blue'))
                if iteration is 'MUGA':
                    tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                                 new_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*\
                           DopplerWind(Temp[i-1], Fre_range_i[xx], Para, wind_v[:,i-1], 
                                       shift_direction='Blue')  # F_vl_i[xx][i-1]
                elif iteration is 'LI' or iteration is 'MALI':
                    tl_2 = basic(Ite_pop[i-1, 0, low_tag, 0],
                                 Ite_pop[i-1, 0, up_tag, 0],
                                 Blu[up_tag, low_tag],
                                 Bul[up_tag, low_tag],
                                 Freq_array[xx]*1.e9)*\
                           DopplerWind(Temp[i-1], Fre_range_i[xx], Para, wind_v[:,i-1], 
                                       shift_direction='Blue')  # F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """======================"""

                Sl = PopuSource_AB(Ite_pop[i, 0, low_tag, 0],
                                   Ite_pop[i, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                if iteration is 'MUGA':
                    Sl1 = PopuSource_AB(new_pop[i-1, 0, low_tag, 0],
                                        new_pop[i-1, 0, up_tag, 0],
                                        Aul[up_tag, low_tag],
                                        Bul[up_tag, low_tag],
                                        Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                elif iteration is 'LI' or iteration is 'MALI':
                    Sl1 = PopuSource_AB(Ite_pop[i-1, 0, low_tag, 0],
                                        Ite_pop[i-1, 0, up_tag, 0],
                                        Aul[up_tag, low_tag],
                                        Bul[up_tag, low_tag],
                                        Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Il = ji_out_all[xx, i-1, :, :]*1.  #
                ji_out_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                          = FOSC(tl1, Sl1, Sl, Il)  # (12.113)
                ji_out_all[xx, i, Mu_tangent > Alt_ref[i], :]=0
                lambda_approx_out[xx, i, Mu_tangent > Alt_ref[i], :] = 0
                ji_out_all[xx, i,
                           Mu_tangent == Alt_ref[i],
                           :] = ji_in_all[xx, i, Mu_tangent == Alt_ref[i], :]
                lambda_approx_out[xx, i,
                           Mu_tangent == Alt_ref[i],
                           :] = lambda_approx_in[xx, i, Mu_tangent == Alt_ref[i], :]
        weighttemp = (mu_weight[:, i].reshape(mu_weight[:, i].size, 1) *
                      np.ones((mu_weight[:, i].size, Fre_range_i[xx].size)))
        weighttemp = weighttemp*1./weighttemp[:, 0].sum()
        for xx in range(Nt):  # transitions
            Para=[Freq_array[xx]*1.e9, 18.0153]
            if wind_v is not None:
                incoming_doppler = DopplerWind(Temp[i],
                                               Fre_range_i[xx],
                                               Para, wind_v[:,i],
                                               shift_direction='Red')
                outgoing_doppler = DopplerWind(Temp[i],
                                               Fre_range_i[xx],
                                               Para, wind_v[:,i],
                                               shift_direction='Blue')
                Fre_weight_i = trapz_inte_edge(incoming_doppler, Fre_range_i[xx])
                Fre_weight_o = trapz_inte_edge(outgoing_doppler, Fre_range_i[xx])
                J_mean = np.sum((ji_out_all[xx, i, :, :] * Fre_weight_o +
                                 ji_in_all[xx, i, :, :] * Fre_weight_i) *
                                weighttemp)*0.5
                l_ap= np.sum((lambda_approx_in[xx, i, :, :] * Fre_weight_i +
                              lambda_approx_out[xx, i, :, :] * Fre_weight_o) *
                             weighttemp)*0.5
#                if (l_ap>1) or (l_ap<0):
#                    print(l_ap, i, xx,
#                          lambda_approx_in[xx, i, :, :].min(),
#                          lambda_approx_out[xx, i, :, :].min(),
#                          lambda_approx_in[xx, i, :, :].max(),
#                          lambda_approx_out[xx, i, :, :].max())
            else: 
                Fre_weight = trapz_inte_edge(F_vl_i[xx][i], Fre_range_i[xx])
                j_mu = np.sum((ji_out_all[xx, i, :, :] +
                              ji_in_all[xx, i, :, :]) *
                              weighttemp,
                              axis=0)*0.5
                J_mean = (j_mu * Fre_weight).sum()  # * FreDelta
                """Lambda operator """
                L_ap= np.sum((lambda_approx_in[xx, i, :, :] +
                              lambda_approx_out[xx, i, :, :]) *
                             weighttemp,
                             axis=0)*0.5
                if (L_ap>1) or (L_ap<0):
                    logger.info(
                        L_ap, i,
                        lambda_approx_in[xx, i, :, :].min(),
                        lambda_approx_out[xx, i, :, :].min(),
                        lambda_approx_in[xx, i, :, :].max(),
                        lambda_approx_out[xx, i, :, :].mmx()
                    )
                    L_ap[L_ap<0]=0
                    L_ap[L_ap>1]=1
                #j_mean_corr[xx,i] = J_mean
                l_ap = (L_ap * Fre_weight).sum()  # * FreDelta
#            l_ap[l_ap<0]=0
#            l_ap[l_ap>1]=1
            if (l_ap>1) or (l_ap<0):
                logger.info(l_ap, i, lambda_approx_in[xx, i, :, :].min(),
                  lambda_approx_out[xx, i, :, :].min())
            B_int[B_place == xx] = J_mean*1.
            if Alt_ref[i] == 0:
                Fre_weight = trapz_inte_edge(F_vl_i[xx][i], Fre_range_i[xx])
                J_mean = (Bv_T(Fre_range_i[xx], Temp[i]) * Fre_weight).sum()
                B_int_lamda[B_place == xx] = J_mean*1.
            elif (i>0) & (i<Alt_ref.size-1) & (iteration is 'MUGA' or iteration is 'MALI'):
                B_int_lamda[B_place == xx] = J_mean*1. - l_ap*SF[xx]
                #print(l_ap*SF[xx])
                #B_int_lamda[B_place == xx] = SF[xx]+(J_mean*1. - SF[xx])/(1.-l_ap)
                A_int_lamda[(B_place == xx)&(RaRaA>0)] = l_ap
                #j_eff_corr[xx,i] = J_mean*1. - l_ap*SF[xx]
            elif (i>0) & (i<Alt_ref.size-1) & (iteration is 'LI'):
                B_int_lamda[B_place == xx] = J_mean*1.
                #print(l_ap*SF[xx])
                #B_int_lamda[B_place == xx] = SF[xx]+(J_mean*1. - SF[xx])/(1.-l_ap)
                A_int_lamda[(B_place == xx)&(RaRaA>0)] = 0.
                #j_eff_corr[xx,i] = J_mean*1. - l_ap*SF[xx]
        RaRij = (RaRaB_absorption+RaRaB_induced) * B_int
        RaRii = np.eye(Ni)*(RaRij.sum(axis=0))*-1.
        A_m = (RaRaA+RaRaAd+RaRij+RaRii+CoRa_block[i])*-1.
        if (i>0) & (i<Alt_ref.size-1) & (iteration is 'MUGA' or iteration is 'MALI'):  # preconditioning part
            RaRij_lambda = (RaRaB_absorption+RaRaB_induced) * B_int_lamda
            RaRii_lambda = np.eye(Ni)*(RaRij_lambda.sum(axis=0))*-1.
            P_m = (RaRaA*(1-A_int_lamda)+(-RaRaA*(1-A_int_lamda)).sum(axis=0)*np.eye(Ni)+RaRij_lambda+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = ((-RaRaA*(1-A_int_lamda)).sum(axis=0)+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = P_m.diagonal()*np.eye(Ni)
            P_m[-1, :] = 1.
        elif (i>0) & (i<Alt_ref.size-1) & (iteration is 'LI'):  # preconditioning part
            RaRij_lambda = (RaRaB_absorption+RaRaB_induced) * B_int_lamda
            RaRii_lambda = np.eye(Ni)*(RaRij_lambda.sum(axis=0))*-1.
            P_m = (RaRaA+(-RaRaA).sum(axis=0)*np.eye(Ni)+RaRij_lambda+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = ((-RaRaA*(1-A_int_lamda)).sum(axis=0)+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = P_m.diagonal()*np.eye(Ni)
            P_m[-1, :] = 1.
        A_m[-1, :] = 1.
        b = np.zeros((Ni, 1))*1.
        b[-1] = Ite_pop[i][0].sum()
        n_old = Ite_pop[i][0]*1.
        #"""correction method (normal?)"""Don't repeat yourself
        #n_delta = b - np.dot(A_m, n_old)
#        if (i>0)&(i<Alt_ref.size-1):  # preconditioning part
#            ##"""method1"""
#            n_new = (np.eye(Ni)-np.linalg.inv(P_m).dot(A_m)).dot(n_old) + np.linalg.inv(P_m).dot(b)
#            ##"""method 2"""
#            #residual_r = b - np.dot(P_m, n_old)
#            #n_delta = np.linalg.inv(P_m).dot(residual_r)
#            #omega = 2./(1+(1-(A_int_lamda**2).max())**0.5)
#            #n_new = n_old + n_delta#*omega  # plus or minus?
#        else:
#            residual_r = b - np.dot(A_m, n_old)
#            n_new = np.linalg.inv(A_m).dot(b)
#            n_delta = n_new-n_old
        # """Inversion method"""
        if (i > 0) & (i < Alt_ref.size-1):  # preconditioning part
            n_new = np.linalg.inv(P_m).dot(b)
        elif i == 0:
            n_new = n_old
        elif i == Alt_ref.size-1:
            n_new = np.linalg.inv(A_m).dot(b)
        else:
            logger.info('check alt. grid', i, Alt_ref[i])
        n_delta = n_new-n_old
        if update_population is True:
            new_pop[i, 0, :] = n_new
        else:
            n_new = n_old
            new_pop[i, 0, :] = n_new #this is for no update
#        print(n_new)
        max_true_error[i] = np.abs((n_delta/n_new)*100).max()
        if (i > 0) and (i < Alt_ref.size-1) and (iteration is 'MUGA'):
            for xx in range(Nt):  # transitions
                Para=[Freq_array[xx]*1.e9, 18.0153]
                Fre_range = Fre_range_i[xx]
                # Line_Shape = F_vl_i[xx]
                Abs_coeff = Abs_ite[xx]
                Fre_range = Fre_range_i[xx]
                up_tag = int(Tran_tag[xx][1])
                low_tag = int(Tran_tag[xx][2])
                # tl_1 = Abs_coeff[i] * F_vl_i[xx][i]
                tl_1 = basic(new_pop[i, 0, low_tag, 0],
                             new_pop[i, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                             new_pop[i-1, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                # tdl2 = calc_abscoeff(xx, IteNum+1, i-1) * F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2).reshape((1, Fre_range.size)) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                tl_3 = Abs_coeff[i+1] * F_vl_i[xx][i+1]
                tl3 = (0.5*np.abs(tl_1+tl_3).reshape((1, Fre_range.size)) *
                       PSC2[:, i].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """===Doppler=========="""
                tl_1 = basic(new_pop[i, 0, low_tag, 0],
                             new_pop[i, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*\
                       DopplerWind(Temp[i], Fre_range_i[xx], Para, wind_v[:,i], shift_direction='Blue')
                tl_2 = basic(new_pop[i-1, 0, low_tag, 0],
                             new_pop[i-1, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*\
                       DopplerWind(Temp[i-1], Fre_range_i[xx], Para, wind_v[:,i-1], 
                                   shift_direction='Blue')  # F_vl_i[xx][i-1]
                tl_3 = Abs_coeff[i+1] * \
                DopplerWind(Temp[i+1], Fre_range_i[xx], Para, wind_v[:,i+1], shift_direction='Red')
                tl1 = (0.5*np.abs(tl_1+tl_2) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                tl3 = (0.5*np.abs(tl_1+tl_3) *
                       PSC2[:, i].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                """======================"""
                Sl2 = PopuSource_AB(new_pop[i, 0, low_tag, 0],
                                    new_pop[i, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Sl1 = PopuSource_AB(new_pop[i-1, 0, low_tag, 0],
                                    new_pop[i-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Sl3 = PopuSource_AB(Ite_pop[i+1, 0, low_tag, 0],
                                    Ite_pop[i+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])
                Il = ji_out_all[xx, i-1, :, :]*1.  # !!!CHECK!!!
                # ji_out_all[xx, i, :, :] = FOSC(tl1, Sl1, Sl2, Il)  # (12.113)
                ji_out_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                          = SOSC(tl1, tl3,
                                 Sl1, Sl2, Sl3,
                                 Il)
                ji_out_all[xx, i, Mu_tangent > Alt_ref[i], :]=0
                lambda_approx_out[xx, i, Mu_tangent > Alt_ref[i], :] = 0
                ji_out_all[xx, i, Mu_tangent == Alt_ref[i], :]\
                = ji_in_all[xx, i, Mu_tangent == Alt_ref[i], :]
                lambda_approx_out[xx, i, Mu_tangent == Alt_ref[i], :]\
                = lambda_approx_in[xx, i, Mu_tangent == Alt_ref[i], :]
    if out_put_spectra is True:
        return new_pop,ji_out_all
    else:
        return new_pop




def jmean_calc(freq_w, mu_w, j_fre_mu):
    j_fre = np.sum(j_fre_mu*mu_w, axis=0)
    j_mean = np.sum(j_fre * freq_w).sum()
    return j_mean


def trans_rate_matrix(trans_place, radiation_coeff, jmean,
                      acoeff_rate, coll_rate):
    j_matrix = trans_place*0.
    for hoge in range(jmean.size):
        j_matrix[trans_place == hoge] = jmean[hoge]
    radiation_rate = radiation_coeff*j_matrix
    radiation_rate_diag = (np.eye(trans_place.shape[0]) *
                           (radiation_rate.sum(axis=0))*-1.)
    trans_matrix = (acoeff_rate +
                    radiation_rate +
                    radiation_rate_diag +
                    coll_rate)*-1.
    return trans_matrix


def pop_calc(trans_rm, old_pop):
    b = np.zeros((trans_rm.shape[0], 1))*1.
    b[-1] = old_pop.sum()
    n_new = np.linalg.inv(trans_rm).dot(b)
    return n_new




def MALI(Ite_pop, Abs_ite,
         PSC2, Mu_tangent, mu_weight,
         Alt_ref, Temp,
         Fre_range_i, Freq_array, F_vl_i,
         B_place,
         Nt, Ni, Aul, Bul, Blu,
         RaRaB_absorption, RaRaB_induced, RaRaA, RaRaAd, CoRa_block,
         Tran_tag):
    ji_in_all = np.zeros((Nt,
                          Alt_ref.size,
                          Mu_tangent.size,
                          Fre_range_i[0].size))
    ji_out_all = ji_in_all*0.
    lambda_approx_in = np.zeros((Nt,
                                 Alt_ref.size,
                                 Mu_tangent.size,
                                 Fre_range_i[0].size))
    lambda_approx_out = np.zeros((Nt,
                                  Alt_ref.size,
                                  Mu_tangent.size,
                                  Fre_range_i[0].size))
    # for IteNum in range(0, 100):
    max_true_error = np.ones(Alt_ref.size)*0.
    new_pop = np.array(Ite_pop)*0.+0.
    for ii in range(Alt_ref.size)[::-1]:  # spatial points
        # print(ii)
        for xx in range(Nt):  # transitions
            # Line_Shape = F_vl_i[xx]
            Abs_coeff = Abs_ite[xx]
            Fre_range = Fre_range_i[xx]
            up_tag = int(Tran_tag[xx][1])
            low_tag = int(Tran_tag[xx][2])
            # Up_pop = Ite_pop[ii, 0, up_tag, 0]
            # Low_pop = Ite_pop[ii, 0, low_tag, 0]
            if ii == Alt_ref.size-1:  # upper Boundary(u>0) incoming
                B_v_cosmic = Bv_T(Fre_range, 2.375)
                ji_in_all[xx, ii, :, :] = 0.*B_v_cosmic  # Ladi comparison
            elif ii == Alt_ref.size-1-1:
                tdu2 = Abs_coeff[ii+1] * F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * F_vl_i[xx][ii]
                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :],lambda_approx_in[xx, ii, :, :]\
                         = FOSC(tdu, Sd1, Sd, Idu)
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
            elif ii == 0:  # SOSC is not available (FOSC elif ii>=0)
                tdu2 = Abs_coeff[ii+1] * F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * F_vl_i[xx][ii]
                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd3 = PopuSource_AB(Ite_pop[ii-1, 0, low_tag, 0],
                                    Ite_pop[ii-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                tdu3 = Abs_coeff[ii-1] * F_vl_i[xx][ii-1]
                tdb = (0.5*np.abs(tdu1+tdu3).reshape((1, Fre_range.size)) *
                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :],lambda_approx_in[xx, ii, :, :]\
                = FOSC(tdu, Sd1, Sd, Idu)
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
            else:  # SOSC
                tdu2 = Abs_coeff[ii+1] * F_vl_i[xx][ii+1]
                tdu1 = Abs_coeff[ii] * F_vl_i[xx][ii]
                tdu = (0.5*np.abs(tdu1+tdu2).reshape((1, Fre_range.size)) *
                       PSC2[:, ii].reshape((Mu_tangent.size, 1)))
                Sd = PopuSource_AB(Ite_pop[ii, 0, low_tag, 0],
                                   Ite_pop[ii, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd1 = PopuSource_AB(Ite_pop[ii+1, 0, low_tag, 0],
                                    Ite_pop[ii+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                Sd3 = PopuSource_AB(Ite_pop[ii-1, 0, low_tag, 0],
                                    Ite_pop[ii-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][ii]
                tdu3 = Abs_coeff[ii-1] * F_vl_i[xx][ii-1]
                tdb = (0.5*np.abs(tdu1+tdu3).reshape((1, Fre_range.size)) *
                       PSC2[:, ii-1].reshape((Mu_tangent.size, 1)))
                Idu = ji_in_all[xx, ii+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, ii, :, :], lambda_approx_in[xx, ii, :, :]\
                         = SOSC(tdu, tdb,
                                Sd1, Sd, Sd3,
                                Idu)
                ji_in_all[xx, ii, (Mu_tangent == Alt_ref[ii]), :],\
                lambda_approx_in[xx, ii, (Mu_tangent == Alt_ref[ii]), :]\
                          = FOSC(tdu[Mu_tangent == Alt_ref[ii]],
                                    Sd1,
                                    Sd,
                                    Idu[Mu_tangent == Alt_ref[ii]])
                ji_in_all[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                lambda_approx_in[xx, ii, ((Mu_tangent > Alt_ref[ii])), :] = 0
                # mu; 0->i,
    for i in range(Alt_ref.size):  # spatial points
        #  print(i)
        B_int = B_place*0.  # intensity for each transitoin
        B_int_lamda = B_place*0.  # intensity for each transitoin for lambda ope
        A_int_lamda = B_place*0.  # approximated lambda operate
        SF = np.ones(Nt)
        for xx in range(Nt):  # transitions
            Fre_range = Fre_range_i[xx]
            # Line_Shape = F_vl_i[xx]
            Abs_coeff = Abs_ite[xx]
            Fre_range = Fre_range_i[xx]
            up_tag = int(Tran_tag[xx][1])
            low_tag = int(Tran_tag[xx][2])
            # Up_pop = Ite_pop[i, 0, up_tag, 0]
            # Low_pop = Ite_pop[i, 0, low_tag, 0]
            if i == 0:  # Lower Boundary(u>0) outgoing
                ji_out_all[xx, i, :, :] = Bv_T(Fre_range, Temp[i])
                ji_out_all[:, i, (Mu_tangent > Alt_ref[i]), :] = 0
            elif i < Alt_ref.size-1:  # SOSC
                tl_1 = Abs_coeff[i] * F_vl_i[xx][i]
                # GS-iteration
                tl_2 = basic(Ite_pop[i-1, 0, low_tag, 0],
                             Ite_pop[i-1, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                # tdl2 = calc_abscoeff(xx, IteNum+1, i-1) * F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2).reshape((1, Fre_range.size)) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                tl_3 = Abs_coeff[i+1] * F_vl_i[xx][i+1]
                tl3 = (0.5*np.abs(tl_1+tl_3).reshape((1, Fre_range.size)) *
                       PSC2[:, i].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                Sl2 = PopuSource_AB(Ite_pop[i, 0, low_tag, 0],
                                    Ite_pop[i, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                SF[xx] = Sl2
                Sl1 = PopuSource_AB(Ite_pop[i-1, 0, low_tag, 0],
                                    Ite_pop[i-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Sl3 = PopuSource_AB(Ite_pop[i+1, 0, low_tag, 0],
                                    Ite_pop[i+1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])
                Il = ji_out_all[xx, i-1, :, :]*1.  #
                # ji_out_all[xx, i, :, :] = FOSC(tl1, Sl1, Sl2, Il)  # (12.113)
                ji_out_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                          = SOSC(tl1, tl3,
                                 Sl1, Sl2, Sl3,
                                 Il)
                ji_out_all[xx, i, Mu_tangent > Alt_ref[i], :]=0
                lambda_approx_out[xx, i, Mu_tangent > Alt_ref[i], :] = 0
                ji_out_all[xx, i, Mu_tangent == Alt_ref[i], :]\
                = ji_in_all[xx, i, Mu_tangent == Alt_ref[i], :]
                lambda_approx_out[xx, i, Mu_tangent == Alt_ref[i], :]\
                = lambda_approx_in[xx, i, Mu_tangent == Alt_ref[i], :]
                I3 = ji_in_all[xx, i+1, :, :]*1.  # !!!CHECK!!!
                ji_in_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                         = SOSC(tl3, tl1,
                                Sl3, Sl2, Sl1,
                                I3)
                ji_in_all[xx, i, (Mu_tangent == Alt_ref[i]), :],\
                lambda_approx_out[xx, i, (Mu_tangent == Alt_ref[i]), :]\
                             = FOSC(tl3[Mu_tangent == Alt_ref[i]],
                                    Sl3,
                                    Sl2,
                                    I3[Mu_tangent == Alt_ref[i]])
                ji_in_all[xx, i, ((Mu_tangent > Alt_ref[i])), :] = 0
                lambda_approx_out[xx, i, ((Mu_tangent > Alt_ref[i])), :] = 0
#                ji_in_all[xx, i, ((Mu_tangent > Alt_ref[i-1])), :] = 0
            elif i == Alt_ref.size-1:  # FOSC
                tl_1 = Abs_coeff[i] * F_vl_i[xx][i]
                # GS-iteration
                tl_2 = basic(Ite_pop[i-1, 0, low_tag, 0],
                             Ite_pop[i-1, 0, up_tag, 0],
                             Blu[up_tag, low_tag],
                             Bul[up_tag, low_tag],
                             Freq_array[xx]*1.e9)*F_vl_i[xx][i-1]
                # tdl2 = calc_abscoeff(xx, IteNum+1, i-1) * F_vl_i[xx][i-1]
                tl1 = (0.5*np.abs(tl_1+tl_2).reshape((1, Fre_range.size)) *
                       PSC2[:, i-1].reshape((Mu_tangent.size, 1)))  # (Mu, Fre)
                Sl = PopuSource_AB(Ite_pop[i, 0, low_tag, 0],
                                   Ite_pop[i, 0, up_tag, 0],
                                   Aul[up_tag, low_tag],
                                   Bul[up_tag, low_tag],
                                   Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Sl1 = PopuSource_AB(Ite_pop[i-1, 0, low_tag, 0],
                                    Ite_pop[i-1, 0, up_tag, 0],
                                    Aul[up_tag, low_tag],
                                    Bul[up_tag, low_tag],
                                    Blu[up_tag, low_tag])  # * F_vl_i[xx][i]
                Il = ji_out_all[xx, i-1, :, :]*1.  #
                ji_out_all[xx, i, :, :],lambda_approx_out[xx, i, :, :]\
                          = FOSC(tl1, Sl1, Sl, Il)  # (12.113)
                ji_out_all[xx, i, Mu_tangent > Alt_ref[i], :]=0
                lambda_approx_out[xx, i, Mu_tangent > Alt_ref[i], :] = 0
                ji_out_all[xx, i,
                           Mu_tangent == Alt_ref[i],
                           :] = ji_in_all[xx, i, Mu_tangent == Alt_ref[i], :]
                lambda_approx_out[xx, i,
                           Mu_tangent == Alt_ref[i],
                           :] = lambda_approx_in[xx, i, Mu_tangent == Alt_ref[i], :]
        for xx in range(Nt):  # transitions
            Fre_weight = trapz_inte_edge(F_vl_i[xx][i], Fre_range_i[xx])
            weighttemp = (mu_weight[:, i].reshape(mu_weight[:, i].size, 1) *
                          np.ones((mu_weight[:, i].size, Fre_weight.size)))
            weighttemp = weighttemp*1./weighttemp[:, 0].sum()
            j_mu = np.sum((ji_out_all[xx, i, :, :] +
                          ji_in_all[xx, i, :, :]) *
                          weighttemp,
                          axis=0)*0.5
            """Lambda operator """
            MALI_LI = 'MALI'
            L_ap= np.sum((lambda_approx_in[xx, i, :, :] +
                          lambda_approx_out[xx, i, :, :]) *
                          weighttemp,
                          axis=0)*0.5
            L_ap[L_ap<0]=0
            J_mean = (j_mu * Fre_weight).sum()  # * FreDelta
            l_ap = (L_ap * Fre_weight).sum()  # * FreDelta
            if Alt_ref[i] == 0:
                J_mean = (Bv_T(Fre_range_i[xx], Temp[i]) * Fre_weight).sum()
            B_int[B_place == xx] = J_mean*1.
            if (i>0) & (i<Alt_ref.size-1):
                B_int_lamda[B_place == xx] = J_mean*1. - l_ap*SF[xx]
                A_int_lamda[(B_place == xx)&(RaRaA>0)] = l_ap
                if MALI_LI=='LI':
                    B_int_lamda[B_place == xx] = J_mean*1.
                    A_int_lamda[(B_place == xx)&(RaRaA>0)] = 0
                #j_eff_corr[xx,i] = J_mean*1. - l_ap*SF[xx]
        RaRij = (RaRaB_absorption+RaRaB_induced) * B_int
        RaRii = np.eye(Ni)*(RaRij.sum(axis=0))*-1.
        A_m = (RaRaA+RaRaAd+RaRij+RaRii+CoRa_block[i])*-1.
        if (i>0) & (i<Alt_ref.size-1):  # preconditioning part
            RaRij_lambda = (RaRaB_absorption+RaRaB_induced) * B_int_lamda
            RaRii_lambda = np.eye(Ni)*(RaRij_lambda.sum(axis=0))*-1.
            P_m = (RaRaA*(1.-A_int_lamda)+(-RaRaA*(1.-A_int_lamda)).sum(axis=0)*np.eye(Ni)+RaRij_lambda+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = ((-RaRaA*(1-A_int_lamda)).sum(axis=0)+RaRii_lambda+CoRa_block[i])*-1.
            #P_m = P_m.diagonal()*np.eye(Ni)
            P_m[-1, :] = 1.
        A_m[-1, :] = 1.
        b = np.zeros((Ni, 1))*1.
        b[-1] = Ite_pop[i][0].sum()
        n_old = Ite_pop[i][0]*1.
        #"""correction method (normal?)"""
        #n_delta = b - np.dot(A_m, n_old)
#        if (i>0)&(i<Alt_ref.size-1):  # preconditioning part
#            ##"""method1"""
#            n_new = (np.eye(Ni)-np.linalg.inv(P_m).dot(A_m)).dot(n_old) + np.linalg.inv(P_m).dot(b)
#            ##"""method 2"""
#            #residual_r = b - np.dot(P_m, n_old)
#            #n_delta = np.linalg.inv(P_m).dot(residual_r)
#            #omega = 2./(1+(1-(A_int_lamda**2).max())**0.5)
#            #n_new = n_old + n_delta#*omega  # plus or minus?
#        else:
#            residual_r = b - np.dot(A_m, n_old)
#            n_new = np.linalg.inv(A_m).dot(b)
#            n_delta = n_new-n_old
        # """Inversion method"""
        if (i>0) & (i<Alt_ref.size-1):  # preconditioning part
            n_new = np.linalg.inv(P_m).dot(b)
        else:
            n_new = np.linalg.inv(A_m).dot(b)
        n_delta = n_new-n_old
        # """Population input """
        new_pop[i, 0, :] = n_new
        max_true_error[i] = np.abs((n_delta/n_new)*100).max()
    return new_pop





"""Implementation of classes to handle various catalogue information. """
import re

import numpy as np
import scipy.sparse

__all__ = ['ArrayOfLineRecord',
           'CIARecord',
           'GasAbsLookup',
           'LineMixingRecord',
           'QuantumIdentifier',
           'QuantumNumberRecord',
           'QuantumNumbers',
           'Sparse',
           'SpeciesAuxData',
           'SpeciesTag',
           'PropagationMatrix',
           'StokesVector',
           'Ppath',
           'GridPos',
           ]


class GridPos:
    """Representation of ARTS GridPos"""
    
    def __init__(self, ind=None, n1=None, n2=None):
        self.ind = ind  # Index
        self.n1 = n1    # Numeric
        self.n2 = n2    # Numeric
        
    @classmethod
    def from_xml(cls, xmlelement):
        """Loads GridPos object from an existing file.
        """
        obj = cls()
        
        obj.ind = int(xmlelement[0].value())
        obj.n1 = xmlelement[1].value()
        obj.n2 = xmlelement[2].value()
        
        return obj
    
    def write_xml(self, xmlwriter, attr=None):
        """Write GridPos object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("GridPos", attr)
        xmlwriter.write_xml(self.ind, {"name": "OriginalGridIndexBelowInterpolationPoint"})
        xmlwriter.write_xml(self.n1, {"name": "FractionalDistanceToNextPoint_1"})
        xmlwriter.write_xml(self.n2, {"name": "FractionalDistanceToNextPoint_2"})
        xmlwriter.close_tag()
    
    def __repr__(self):
        return "GridPos {}: n1={}; n2={}".format(self.ind, self.n1, self.n2)
    
    def __eq__(self, other):
        return self.ind == other.ind and self.n1 == other.n1 and self.n2 == other.n2
    
    def __ne__(self, other):
        return not (self==other)


class Ppath:
    """Represents the Ppath variable in ARTS"""
    
    def __init__(self):
        self.dim = None          # Index
        self.np = None           # Index
        self.constant = None     # Numeric
        self.background = None   # String
        self.start_pos = None    # Vector
        self.start_los = None    # Vector
        self.start_lstep = None  # Numeric
        self.pos = None          # Matrix
        self.los = None          # Matrix
        self.r = None            # Vector
        self.lstep = None        # Vector
        self.end_pos = None      # Vector
        self.end_los = None      # Vector
        self.end_lstep = None    # Vector
        self.nreal = None        # Vector
        self.ngroup = None       # Vector
        self.gp_p = None         # ArrayOfGridPos
        self.gp_lat = None       # ArrayOfGridPos
        self.gp_lon = None       # ArrayOfGridPos
    
    def __repr__(self):
        return "Ppath of {} steps in {}D Atmosphere".format(self.np, self.dim)
    
    @classmethod
    def from_xml(cls, xmlelement):
        """Loads Ppath object from an existing file.
        """
        obj = cls()
        
        obj.dim = int(xmlelement[0].value())
        obj.np = int(xmlelement[1].value())
        obj.constant = xmlelement[2].value()
        obj.background = xmlelement[3].value()
        obj.start_pos = xmlelement[4].value()
        obj.start_los = xmlelement[5].value()
        obj.start_lstep = xmlelement[6].value()
        obj.pos = xmlelement[7].value()
        obj.los = xmlelement[8].value()
        obj.r = xmlelement[9].value()
        obj.lstep = xmlelement[10].value()
        obj.end_pos = xmlelement[11].value()
        obj.end_los = xmlelement[12].value()
        obj.end_lstep = xmlelement[13].value()
        obj.nreal = xmlelement[14].value()
        obj.ngroup = xmlelement[15].value()
        obj.gp_p = xmlelement[16].value()
        obj.gp_lat = xmlelement[17].value()
        obj.gp_lon = xmlelement[18].value()
        return obj
    
    def write_xml(self, xmlwriter, attr=None):
        """Write Ppath object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("Ppath", attr)
        xmlwriter.write_xml(self.dim, {"name": "AtmosphericDimensionality"})
        xmlwriter.write_xml(self.np, {"name": "NumberOfPositionInPropagationPath"})
        xmlwriter.write_xml(self.constant, {"name": "PropagationPathConstant"})
        xmlwriter.write_xml(self.background, {"name": "RadiativeBackground"})
        xmlwriter.write_xml(self.start_pos, {"name": "StartPositionOfPropagationPath"})
        xmlwriter.write_xml(self.start_los, {"name": "StartLOSOfPropagationPath"})
        xmlwriter.write_xml(self.start_lstep, {"name": "StartLstepOfPropagationPath"})
        xmlwriter.write_xml(self.pos, {"name": "PropagationPathPointPositions"})
        xmlwriter.write_xml(self.los, {"name": "LineOfSight"})
        xmlwriter.write_xml(self.r, {"name": "PropagationPathPointRadii"})
        xmlwriter.write_xml(self.lstep, {"name": "PropagationPathPositionLength"})
        xmlwriter.write_xml(self.end_pos, {"name": "EndPositionOfPropagationPath"})
        xmlwriter.write_xml(self.end_los, {"name": "EndLOSOfPropagationPath"})
        xmlwriter.write_xml(self.end_lstep, {"name": "EndLstepPropagationPath"})
        xmlwriter.write_xml(self.nreal, {"name": "RefractiveIndexRealPart"})
        xmlwriter.write_xml(self.ngroup, {"name": "GroupRefractiveIndex"})
        xmlwriter.write_xml(self.gp_p, {"name": "PressureGridIndexPosition"}, arraytype="GridPos")
        xmlwriter.write_xml(self.gp_lat, {"name": "LatitudeGridIndexPosition"}, arraytype="GridPos")
        xmlwriter.write_xml(self.gp_lon, {"name": "LongitudeGridIndexPosition"}, arraytype="GridPos")
        xmlwriter.close_tag()
        
    def alt_lat_lon_za_aa(self):
        alt = self.pos[:, 0]
        lat = self.pos[:, 1] if self.pos.shape[1] > 1 else np.zeros_like(alt)
        lon = self.pos[:, 2] if self.pos.shape[1] > 2 else np.zeros_like(alt)
        
        za = self.los[:, 0]
        aa = self.los[:, 1] if self.los.shape[1] > 1 else np.zeros_like(za)
        
        return alt, lat, lon, za, aa


class ArrayOfLineRecord:
    """Represents an :arts:`ArrayOfLineRecord` object."""

    def __init__(self, data=None, version=None):
        self.data = data
        self.version = version

    def __repr__(self):
        if len(self.data) > 1:
            return "ArrayOfLineRecord. " + self.version + ". " + \
                str(len(self.data)) + " lines."
        elif len(self.data) == 1:
            if '@' in self.data[0]:
                return "ArrayOfLineRecord. " + self.version + ". 1 line."
        return "ArrayOfLineRecord. " + self.version + ". No lines."

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

    def as_ARTSCAT5(self):
        """Returns manipulable ARTSCAT5 class of this linerecord array
        """
        assert self.version == 'ARTSCAT-5', "Only for ARTSCAT-5 data"
        return ARTSCAT5(self)

    @property
    def version(self):
        """ArrayOfRecord version number."""
        return self._version

    @property
    def data(self):
        """List of strings representing line records."""
        return self._data

    @version.setter
    def version(self, version):
        if version is None:
            self._version = None
            return

        if isinstance(version, str):
            self._version = version
        else:
            raise TypeError('version has to be String.')

    @data.setter
    def data(self, data):
        self._data = return_if_arts_type(data, 'ArrayOfString')

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads an ArrayOfLineRecord object from an existing file.
        """
        obj = cls()
        obj.version = xmlelement.attrib['version']
        obj.data = xmlelement.text.strip().split('\n')
        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write an ArrayOfLineRecord object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        attr['version'] = self.version
        attr['nelem'] = len(self.data)

        xmlwriter.open_tag("ArrayOfLineRecord", attr)
        xmlwriter.write('\n'.join(self.data) + '\n')
        xmlwriter.close_tag()


class CIARecord:
    """Represents a CIARecord object.

    See online ARTS documentation for object details.

    """

    def __init__(self, molecule1=None, molecule2=None, data=None):
        self.molecule1 = molecule1
        self.molecule2 = molecule2
        self.data = data

    @property
    def molecule1(self):
        """Name of the first molecule."""
        return self._molecule1

    @property
    def molecule2(self):
        """Name of the second molecule."""
        return self._molecule2

    @property
    def data(self):
        """Actual data stored in (list of) GriddedField2 objects."""
        return self._data

    @molecule1.setter
    def molecule1(self, molecule1):
        self._molecule1 = return_if_arts_type(molecule1, 'String')

    @molecule2.setter
    def molecule2(self, molecule2):
        self._molecule2 = return_if_arts_type(molecule2, 'String')

    @data.setter
    def data(self, data):
        self._data = return_if_arts_type(data, 'ArrayOfGriddedField2')

    def __repr__(self):
        return self._molecule1 + "-CIA-" + self.molecule2 + " " + \
            str(self.data)

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a CIARecord object from an existing file.
        """

        obj = cls()
        obj.molecule1 = xmlelement.attrib['molecule1']
        obj.molecule2 = xmlelement.attrib['molecule2']
        obj.data = xmlelement[0].value()

        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a CIARecord object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        attr['molecule1'] = self.molecule1
        attr['molecule2'] = self.molecule2

        xmlwriter.open_tag("CIARecord", attr)
        xmlwriter.write_xml(self.data)
        xmlwriter.close_tag()


# TODO(LKL): consider splitting SpeciesAuxData into seperate classes for each
# version. SpeciesAuxData could be used as wrapper class.
class SpeciesAuxData:
    """Represents a SpeciesAuxData object.

    See online ARTS documentation for object details.

    """

    def __init__(self, data, version, nparam=None):
        self.version = version
        self.nparam = nparam
        self.data = data

    def __repr__(self):
        return "SpeciesAuxData Version " + str(self.version) + ' ' + \
            'for ' + str(len(self.species())) + ' species'

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data
        if self.version == 1:
            self._data_dict = {}
            self._keys = {}
            for ii in range(len(data)):
                iso_data = data[ii]
                tmp = iso_data.split()
                self._keys[tmp[1]] = ii
                self._data_dict[tmp[1]] = float(tmp[2])
        elif self.version == 2:
            self._data_dict = {}
            self._keys = {}
            for ii in range(len(data)):
                tmp = data[ii]
                self._keys[tmp[0]] = ii
                self._data_dict[tmp[0]] = [tmp[1], tmp[2]]

    def __getitem__(self, key):
        return self._data_dict[key]

    def __setitem__(self, key, val):
        self._data_dict[key] = val
        if self.version == 1:
            self._data[(self._keys[key])] = '@ ' + key + ' ' + str(val)
        elif self.version == 2:
            self._data[(self._keys[key])] = val

    def __contains__(self, key):
        return key in self._data_dict

    def species(self):
        return list(self._data_dict.keys())

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a SpeciesAuxData object from an existing file.
        """

        version = int(xmlelement.attrib['version'])

        if version == 1:
            nparam = int(xmlelement.attrib['nparam'])
            data = [s for s in xmlelement.text.split('\n') if s != '']
        elif version == 2:
            nparam = None
            data = []
            sub_list = []
            for n, elem in enumerate(xmlelement):
                if n != 0 and n % 3 == 0:
                    data.append(sub_list)
                    sub_list = []
                sub_list.append(elem.value())
            data.append(sub_list)
        else:
            raise Exception(
                "Unknown SpeciesAuxData version {}.".format(version))

        obj = cls(data, version, nparam=nparam)
        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a ScatterinMetaData object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        attr['version'] = self.version
        attr['nelem'] = len(self.data)

        if self.version == 1:
            attr['nparam'] = self.nparam

            xmlwriter.open_tag("SpeciesAuxData", attr)
            xmlwriter.write('\n'.join(self.data) + '\n')
            xmlwriter.close_tag()
        elif self.version == 2:
            xmlwriter.open_tag("SpeciesAuxData", attr)
            for sub_list in self.data:
                for element in sub_list:
                    xmlwriter.write_xml(element)
            xmlwriter.close_tag()

    def as_PartitionFunctions(self):
        return PartitionFunctions(self)


class GasAbsLookup:
    """Represents a GasAbsLookup object.

    See online ARTS documentation for object details.

    """

    def __init__(self,
                 speciestags=None,
                 nonlinearspecies=None,
                 frequencygrid=None,
                 pressuregrid=None,
                 referencevmrprofiles=None,
                 referencetemperatureprofile=None,
                 temperatureperturbations=None,
                 nonlinearspeciesvmrperturbations=None,
                 absorptioncrosssection=None):

        self.speciestags = speciestags
        self.nonlinearspecies = nonlinearspecies
        self.frequencygrid = frequencygrid
        self.pressuregrid = pressuregrid
        self.referencevmrprofiles = referencevmrprofiles
        self.referencetemperatureprofile = referencetemperatureprofile
        self.temperatureperturbations = temperatureperturbations
        self.nonlinearspeciesvmrperturbations = nonlinearspeciesvmrperturbations
        self.absorptioncrosssection = absorptioncrosssection

    @property
    def speciestags(self):
        """List of :class:`SpeciesTag`."""
        return self._speciestags

    @property
    def nonlinearspecies(self):
        """Indices to indentify nonlinear species."""
        return self._nonlinearspecies

    @property
    def frequencygrid(self):
        """Frequency vector."""
        return self._frequencygrid

    @property
    def pressuregrid(self):
        """Pressure level vector."""
        return self._pressuregrid

    @property
    def referencevmrprofiles(self):
        """Reference VMR profiles."""
        return self._referencevmrprofiles

    @property
    def referencetemperatureprofile(self):
        """Reference temperature profile."""
        return self._referencetemperatureprofile

    @property
    def temperatureperturbations(self):
        """Vector with temperature perturbations."""
        return self._temperatureperturbations

    @property
    def nonlinearspeciesvmrperturbations(self):
        """Vector with VMR perturbations for nonlinear species."""
        return self._nonlinearspeciesvmrperturbations

    @property
    def absorptioncrosssection(self):
        """Absorption crosssections."""
        return self._absorptioncrosssection

    @speciestags.setter
    def speciestags(self, speciestags):
        self._speciestags = return_if_arts_type(
            speciestags, 'ArrayOfArrayOfSpeciesTag')

    @nonlinearspecies.setter
    def nonlinearspecies(self, nonlinearspecies):
        self._nonlinearspecies = return_if_arts_type(
            nonlinearspecies, 'ArrayOfIndex')

    @frequencygrid.setter
    def frequencygrid(self, frequencygrid):
        self._frequencygrid = return_if_arts_type(
            frequencygrid, 'Vector')

    @pressuregrid.setter
    def pressuregrid(self, pressuregrid):
        self._pressuregrid = return_if_arts_type(
            pressuregrid, 'Vector')

    @referencevmrprofiles.setter
    def referencevmrprofiles(self, referencevmrprofiles):
        self._referencevmrprofiles = return_if_arts_type(
            referencevmrprofiles, 'Matrix')

    @referencetemperatureprofile.setter
    def referencetemperatureprofile(self, referencetemperatureprofile):
        self._referencetemperatureprofile = return_if_arts_type(
            referencetemperatureprofile, 'Vector')

    @temperatureperturbations.setter
    def temperatureperturbations(self, temperatureperturbations):
        self._temperatureperturbations = return_if_arts_type(
            temperatureperturbations, 'Vector')

    @nonlinearspeciesvmrperturbations.setter
    def nonlinearspeciesvmrperturbations(self, nonlinearspeciesvmrperturbations):
        self._nonlinearspeciesvmrperturbations = return_if_arts_type(
            nonlinearspeciesvmrperturbations, 'Vector')

    @absorptioncrosssection.setter
    def absorptioncrosssection(self, absorptioncrosssection):
        self._absorptioncrosssection = return_if_arts_type(
            absorptioncrosssection, 'Tensor4')

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a GasAbsLookup object from an existing file.
        """

        obj = cls()
        obj.speciestags = xmlelement[0].value()
        obj.nonlinearspecies = xmlelement[1].value()
        obj.frequencygrid = xmlelement[2].value()
        obj.pressuregrid = xmlelement[3].value()
        obj.referencevmrprofiles = xmlelement[4].value()
        obj.referencetemperatureprofile = xmlelement[5].value()
        obj.temperatureperturbations = xmlelement[6].value()
        obj.nonlinearspeciesvmrperturbations = xmlelement[7].value()
        obj.absorptioncrosssection = xmlelement[8].value()

        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a ScatterinMetaData object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("GasAbsLookup", attr)
        xmlwriter.write_xml(self.speciestags)
        if self.nonlinearspecies is None:
            nonlinearspecies = []
        else:
            nonlinearspecies = self.nonlinearspecies
        xmlwriter.write_xml(nonlinearspecies,
                            {'name': 'NonlinearSpecies'},
                            arraytype='Index')
        xmlwriter.write_xml(self.frequencygrid,
                            {'name': 'FrequencyGrid'})
        xmlwriter.write_xml(self.pressuregrid,
                            {'name': 'PressureGrid'})
        xmlwriter.write_xml(self.referencevmrprofiles,
                            {'name': 'ReferenceVmrProfiles'})
        xmlwriter.write_xml(self.referencetemperatureprofile,
                            {'name': 'ReferenceTemperatureProfile'})
        xmlwriter.write_xml(self.temperatureperturbations,
                            {'name': 'TemperaturePerturbations'})
        xmlwriter.write_xml(self.nonlinearspeciesvmrperturbations,
                            {'name': 'NonlinearSpeciesVmrPerturbations'})
        xmlwriter.write_xml(self.absorptioncrosssection,
                            {'name': 'AbsorptionsCrossSections'})
        xmlwriter.close_tag()


class SpeciesTag(str):
    """Represents a SpeciesTag object.

    See online ARTS documentation for object details.

    """

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a SpeciesTag object from an existing file.
        """
        if xmlelement.text is None:
            raise Exception('SpeciesTag must not be empty.')
        return cls(xmlelement.text.strip()[1:-1])

    def write_xml(self, xmlwriter, attr=None):
        """Write a SpeciesTag object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag('SpeciesTag', attr, newline=False)
        xmlwriter.write('"' + self + '"')
        xmlwriter.close_tag()

    def to_tex(self):
        """Turn indices in species tags to LaTeX subscripts."""
        if "FC" in self:
            return self
        else:
            return re.sub("([a-zA-Z]+)([0-9]+)", r"\1$_{\2}$", self)


class Sparse(scipy.sparse.csc_matrix):
    """Wrapper around :class:`scipy.sparse.csc_matrix`.

    This class wraps around the SciPy Compressed Sparse Column matrix. The
    usage is exactly the same, but support for reading and writing XML files
    is added. Also additional attributes were added to map the ARTS
    implementation of :arts:`Sparse`.

    """
    @property
    def nrows(self):
        """Number of rows."""
        return self.shape[0]

    @property
    def ncols(self):
        """Number of columns."""
        return self.shape[0]

    @property
    def rowindex(self):
        """Row indices to locate data in matrix."""
        return self.tocoo().row

    @property
    def colindex(self):
        """Column indices to locate data in matrix."""
        return self.tocoo().col

    @property
    def sparsedata(self):
        """Data value at specified positions in matrix."""
        return self.tocoo().data

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a Sparse object from an existing file."""

        binaryfp = xmlelement.binaryfp
        nelem = int(xmlelement[0].attrib['nelem'])
        nrows = int(xmlelement.attrib['nrows'])
        ncols = int(xmlelement.attrib['ncols'])

        if binaryfp is None:
            rowindex = np.fromstring(xmlelement[0].text, sep=' ').astype(int)
            colindex = np.fromstring(xmlelement[1].text, sep=' ').astype(int)
            sparsedata = np.fromstring(xmlelement[2].text, sep=' ')
        else:
            rowindex = np.fromfile(binaryfp, dtype='<i4', count=nelem)
            colindex = np.fromfile(binaryfp, dtype='<i4', count=nelem)
            sparsedata = np.fromfile(binaryfp, dtype='<d', count=nelem)

        return cls((sparsedata, (rowindex, colindex)), [nrows, ncols])

    def write_xml(self, xmlwriter, attr=None):
        """Write a Sparse object to an ARTS XML file."""

        # Get ARTS-style information from CSC matrix.
        nrows = self.shape[0]
        ncols = self.shape[1]
        rowindex = self.tocoo().row
        colindex = self.tocoo().col
        sparsedata = self.tocoo().data

        precision = xmlwriter.precision

        if attr is None:
            attr = {}

        attr['nrows'] = nrows
        attr['ncols'] = ncols

        xmlwriter.open_tag('Sparse', attr)

        binaryfp = xmlwriter.binaryfilepointer

        if binaryfp is None:
            xmlwriter.open_tag('RowIndex', {'nelem': rowindex.size})
            for i in rowindex:
                xmlwriter.write('%d' % i + '\n')
            xmlwriter.close_tag()
            xmlwriter.open_tag('ColIndex', {'nelem': colindex.size})
            for i in colindex:
                xmlwriter.write('%d' % i + '\n')
            xmlwriter.close_tag()
            xmlwriter.open_tag('SparseData', {'nelem': sparsedata.size})
            for i in sparsedata:
                xmlwriter.write(('%' + precision) % i + '\n')
            xmlwriter.close_tag()
            xmlwriter.close_tag()
        else:
            xmlwriter.open_tag('RowIndex', {'nelem': rowindex.size})
            np.array(rowindex, dtype='i4').tofile(binaryfp)
            xmlwriter.close_tag()
            xmlwriter.open_tag('ColIndex', {'nelem': colindex.size})
            np.array(colindex, dtype='i4').tofile(binaryfp)
            xmlwriter.close_tag()
            xmlwriter.open_tag('SparseData', {'nelem': sparsedata.size})
            np.array(sparsedata, dtype='d').tofile(binaryfp)
            xmlwriter.close_tag()
            xmlwriter.close_tag()


class QuantumIdentifier:
    """Represents a QuantumIdentifier object.

    See online ARTS documentation for object details.

    """

    def __init__(self, qid):

        assert type(qid) is str, "Need String input"
        these = qid.split()
        assert len(these) > 0, "No QuantumIdentifier"
        spec = these[0].split('-')  # UGLY: What about negative charge?
        if len(spec) == 1:
            self._afgl = None
            if spec[0] == 'None':
                self._spec = None
            else:
                self._spec = spec[0]
        elif len(spec) == 2:
            self._spec = spec[0]
            self._afgl = int(spec[1])
        else:
            assert False, "Cannot recognize species"

        if len(these) == 1:
            self._transition = False
            self._level = False
            return

        if these[1] == 'TR':
            self._transition = True
            self._level = False
        elif these[1] == 'EN':
            self._transition = False
            self._level = True
        else:
            assert False, "Must be energy level [EN] or transition [TR] type"

        self._qns = as_quantumnumbers(" ".join(these[2:]))

        self._assert_sanity_()

    def __repr__(self):
        out = str(self._spec)
        if self._afgl is not None:
            out += '-' + str(self._afgl)
        if self._transition or self._level:
            if self._transition:
                out += ' TR '
            else:
                out += ' EN '
            out += str(self._qns)
        return out

    def _assert_sanity_(self):
        if self._transition:
            assert type(self._qns) is QuantumNumberRecord, "Mismatching types"
        elif self._level:
            assert type(self._qns) is QuantumNumbers, "Mismatching types"
        else:
            assert False, "Programmer error?"

    def __str__(self):
        assert self.afgl is not None or self.species is not None, \
            "Bad data cannot be converted to str.  Contains no species or iso"
        return self.__repr__()

    @property
    def qns(self):
        return self._qns

    @qns.setter
    def qns(self, qns):
        self._qns = as_quantumnumbers(qns)
        if type(self._qns) is QuantumNumberRecord:
            self._transition = True
            self._level = False
        elif type(self._qns) is QuantumNumbers:
            self._transition = False
            self._level = True
        else:
            assert False, "Programmer error?"

    @property
    def species(self):
        return self._spec

    @species.setter
    def species(self, value):
        self._spec = return_if_arts_type(value, 'String')

    @property
    def afgl(self):
        return self._afgl

    @afgl.setter
    def afgl(self, value):
        self._afgl = return_if_arts_type(value, 'Index')

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a QuantumIdentifier object from an existing file.
        """
        if xmlelement.text is None:
            raise Exception('QuantumIdentifier must not be empty.')
        return cls(xmlelement.text.strip())

    def write_xml(self, xmlwriter, attr=None):
        """Write a QuantumIdentifier object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag('QuantumIdentifier', attr, newline=False)
        xmlwriter.write(self.__str__())
        xmlwriter.close_tag()


class QuantumNumberRecord:
    """Represents a QuantumNumberRecord object.

    See online ARTS documentation for object details.

    """

    def __init__(self, upper=None, lower=None):
        self._qns = {'UP': QuantumNumbers(), 'LO': QuantumNumbers()}
        self._qns['UP'] = return_if_arts_type(upper, 'QuantumNumbers')
        self._qns['LO'] = return_if_arts_type(lower, 'QuantumNumbers')

    def __repr__(self):
        if len(self._qns['UP']) == 0 and len(self._qns['LO']) == 0:
            return 'No Quantum-Numbers'
        else:
            return "UP " + str(self._qns['UP']) + " LO " + str(self._qns['LO'])

    def __str__(self):
        if len(self._qns['UP']) == 0 and len(self._qns['LO']) == 0:
            return ''
        else:
            return self.__repr__()

    def __getitem__(self, key):
        return self._qns[key]

    def __setitem__(self, key, value):
        self._qns[key] = return_if_arts_type(as_quantumnumbers(value),
                                             'QuantumNumbers')

    def __iter__(self):
        return iter(self._qns)

    def __contains__(self, value):
        return value in ['UP', 'LO']

    @classmethod
    def from_dict(cls, d):
        """Creates a QuantumNumberRecord from dictionary
        """
        if len(d) == 0:
            return QuantumNumberRecord(upper=QuantumNumbers(),
                                       lower=QuantumNumbers())

        assert 'UP' in d and 'LO' in d, "Need UP and LO to create"
        qnr = QuantumNumberRecord(upper=QuantumNumbers(d['UP']),
                                  lower=QuantumNumbers(d['LO']))
        return qnr

    @classmethod
    def from_str(cls, s):
        """Creates a QuantumNumberRecord from string
        """
        s = s.strip()
        if len(s) == 0:
            return QuantumNumberRecord(upper=QuantumNumbers(),
                                       lower=QuantumNumbers())

        assert 'UP' in s and 'LO' in s, "Need UP and LO to create"
        _t1 = s.split('UP')
        assert len(_t1) == 2, "Unexpectedly many/few UP in s"
        if len(_t1[0]) == 0:
            _t2 = _t1[1].split('LO')
            assert len(_t2) == 2, "Unexpectedly many/few LO in s"
            lo = _t2[1]
            up = _t2[0]
        else:
            up = _t1[1]
            _t2 = _t1[0].split('LO')
            assert len(_t2) == 2, "Unexpectedly many/few LO in s"
            lo = _t2[1]

        qnr = QuantumNumberRecord(upper=QuantumNumbers(up),
                                  lower=QuantumNumbers(lo))
        return qnr

    @property
    def upper(self):
        """QuantumNumbers object representing the upper quantumnumber."""
        return self._qns['UP']

    @property
    def lower(self):
        """QuantumNumbers object representing the lower quantumnumber."""
        return self._qns['LO']

    @upper.setter
    def upper(self, upper):
        self._qns['UP'] = return_if_arts_type(upper, 'QuantumNumbers')

    @lower.setter
    def lower(self, lower):
        self._qns['LO'] = return_if_arts_type(lower, 'QuantumNumbers')

    @property
    def qns(self):
        return self._qns

    @qns.setter
    def qns(self, value):
        if 'LO' in value:
            self._qns['LO'] = QuantumNumbers(value['LO'])
        else:
            self._qns['LO'] = QuantumNumbers()

        if 'UP' in value:
            self._qns['UP'] = QuantumNumbers(value['UP'])
        else:
            self._qns['UP'] = QuantumNumbers()

    def zeeman_splitting(self, type=None, case=None, H=1):
        from ..physics.em import zeeman_splitting, landau_g_factor, \
            zeeman_transitions

        if case.lower() == 'a':
            gu = landau_g_factor(self.upper['Omega'], self.upper['J'],
                                 self.upper['S'], self.upper['Lambda'],
                                 gs=2, gl=1, case='a')
            gl = landau_g_factor(self.lower['Omega'], self.lower['J'],
                                 self.lower['S'], self.lower['Lambda'],
                                 gs=2, gl=1, case='a')
        elif case.lower() == 'b':
            gu = landau_g_factor(self.upper['N'], self.upper['J'],
                                 self.upper['S'], self.upper['Lambda'],
                                 gs=2, gl=1, case='b')
            gl = landau_g_factor(self.lower['N'], self.lower['J'],
                                 self.lower['S'], self.lower['Lambda'],
                                 gs=2, gl=1, case='b')
        else:
            raise RuntimeError("No unknown cases allowed")

        if type is not None:
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], type)
            return zeeman_splitting(gu, gl, mu, ml, H)
        else:
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "PI")
            pi = zeeman_splitting(gu, gl, mu, ml, H)
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "S+")
            sp = zeeman_splitting(gu, gl, mu, ml, H)
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "S-")
            sm = zeeman_splitting(gu, gl, mu, ml, H)
            return {"pi": pi, "s+": sp, "s-": sm}

    def zeeman_transitions(self, type=None):
        from ..physics.em import zeeman_transitions, zeeman_strength
        if type is not None:
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], type)
            return zeeman_strength(self.upper['J'], self.lower['J'], mu, ml)
        else:
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "PI")
            pi = zeeman_strength(self.upper['J'], self.lower['J'], mu, ml)
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "S+")
            sp = zeeman_strength(self.upper['J'], self.lower['J'], mu, ml)
            mu, ml = zeeman_transitions(self.upper['J'], self.lower['J'], "S-")
            sm = zeeman_strength(self.upper['J'], self.lower['J'], mu, ml)
            return {"pi": pi, "s+": sp, "s-": sm}

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a QuantumNumberRecord object from an existing file.
        """

        obj = cls()
        obj.upper = xmlelement[0][0].value()
        obj.lower = xmlelement[1][0].value()

        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a SpeciesTag object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag('QuantumNumberRecord', attr)
        xmlwriter.open_tag('Upper', attr, newline=False)
        xmlwriter.write_xml(self.upper)
        xmlwriter.close_tag()
        xmlwriter.open_tag('Lower', attr, newline=False)
        xmlwriter.write_xml(self.lower)
        xmlwriter.close_tag()
        xmlwriter.close_tag()

    def __iadd__(self, qnr):
        self._qns['UP'] += qnr['UP']
        self._qns['LO'] += qnr['LO']

    def __isub__(self, qnr):
        self._qns['UP'] -= qnr['UP']
        self._qns['LO'] -= qnr['LO']

    def __eq__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] == qns['LO'] and self['UP'] == qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] == qns, self['LO'] == qns
        else:
            return self == as_quantumnumbers(qns)

    def __ne__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] != qns['LO'] and self['UP'] != qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] != qns, self['LO'] != qns
        else:
            return self != as_quantumnumbers(qns)

    def __lt__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] < qns['LO'] and self['UP'] < qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] < qns, self['LO'] < qns
        else:
            return self < as_quantumnumbers(qns)

    def __gt__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] > qns['LO'] and self['UP'] > qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] > qns, self['LO'] > qns
        else:
            return self > as_quantumnumbers(qns)

    def __le__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] <= qns['LO'] and self['UP'] <= qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] <= qns, self['LO'] <= qns
        else:
            return self <= as_quantumnumbers(qns)

    def __ge__(self, qns):
        if type(qns) is QuantumNumberRecord:
            return self['LO'] >= qns['LO'] and self['UP'] >= qns['UP']
        elif type(qns) is QuantumNumbers:
            return self['UP'] >= qns, self['LO'] >= qns
        else:
            return self >= as_quantumnumbers(qns)


class QuantumNumbers:
    """Represents a QuantumNumbers object.

    See online ARTS documentation for object details.

    """

    def __init__(self, numbers=None, nelem=None):

        self.numbers = numbers
        if nelem is not None:
            self.nelem = nelem
        else:
            self.nelem = len(self.numbers)

        self._assert_sanity_()

    def _assert_sanity_(self):
        if self.nelem is None or self.numbers is None:
            return
        assert len(self.numbers) == self.nelem, "mismatching quantum numbers"

    def __repr__(self):
        out = ''
        for qn in self.numbers:
            out += qn + ' ' + str(self.numbers[qn]) + ' '
        return out[:-1]

    def __getitem__(self, key):
        """Returns the value.  Mimics ARTS' behavior for mismatched data
        """
        if key in self:
            return self.numbers[key]
        else:
            return None

    def __setitem__(self, key, value):
        """Sets a value and counts up the quantum numbers
        """
        if key in self.numbers:
            self.numbers[key] = Rational(value)
        else:
            self.numbers[key] = Rational(value)
            self.nelem += 1
        self._assert_sanity_()

    def __iadd__(self, qns):
        for qn in qns:
            assert qn not in self, "Addition means adding new QN. Access " + \
                "individual elements to change their values"
            self.numbers[qn] = qns[qn]
            self.nelem += 1
        return self

    def __isub__(self, qns):
        for qn in qns:
            assert qn in self, "Subtraction means removing QN. Access " + \
                "individual elements to change their values"
            del self.numbers[qn]
            self.nelem -= 1
        return self

    def __contains__(self, key):
        """Are these quantum numbers here?
        """
        return key in self.numbers

    def __iter__(self):
        return iter(self.numbers)

    def __eq__(self, qns):
        """Tests for complete equality ==
        """
        return self <= qns and len(qns) == self.nelem

    def __ne__(self, qns):
        """Tests for lacking complete equality !=
        """
        return not self == qns

    def __le__(self, qns):
        """Tests for all in self being in qns <=
        """
        try:
            for qn in self:
                if qns[qn] != self[qn]:
                    return False
            return True
        except:
            return False

    def __ge__(self, qns):
        """Tests for all in qns being in self >=
        """
        try:
            for qn in qns:
                if qns[qn] != self[qn]:
                    return False
            return True
        except:
            return False

    def __lt__(self, qns):
        """Tests for all in self being in qns and if there is more in qns <
        """
        return self <= qns and self.nelem < len(qns)

    def __gt__(self, qns):
        """Tests for all in self being in qns and if there is more in self >
        """
        return qns <= self and len(qns) < self.nelem

    def __len__(self):
        return self.nelem

    def array_of_M(self):
        """Returns all possible M in a list.  Requires presence of J
        """
        assert 'J' in self, "Need J to define M"
        assert self['J'] >= 0, "Negative J in this instance?"
        _t = []
        _s = -self['J']
        while _s <= self['J']:
            _t.append(_s)
            _s += 1
        return np.array(_t)

    @property
    def numbers(self):
        """Dict representing the quantumnumbers."""
        return self._numbers

    @property
    def nelem(self):
        """Number of quantumnumbers stored."""
        return self._nelem

    @numbers.setter
    def numbers(self, numbers):
        if type(numbers) is str:
            _t = numbers.split()
            nums = {}
            i = 0
            assert len(_t) % 2 == 0, "Not of form 'key1 value1 key2 value2'"
            while i < len(_t):
                nums[_t[i]] = Rational(_t[i+1])
                i += 2
            self._numbers = nums
        elif type(numbers) is dict:
            for i in numbers:
                numbers[i] = Rational(numbers[i])
            self._numbers = numbers
        elif type(numbers) is QuantumNumbers:
            self._numbers = numbers.numbers
        elif numbers is None:
            self._numbers = {}
        else:
            assert False, "Expected dict or String for QuantumNumbers"
        # OLD: self._numbers = return_if_arts_type(numbers, 'String')

    @nelem.setter
    def nelem(self, nelem):
        if nelem is None:
            self._nelem = None
            return

        self._nelem = nelem

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a QuantumNumbers object from an existing file.
        """

        obj = cls()
        obj.numbers = xmlelement.text
        obj.nelem = int(xmlelement.attrib['nelem'])

        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a SpeciesTag object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        attr['nelem'] = self.nelem

        xmlwriter.open_tag('QuantumNumbers', attr, newline=False)
        xmlwriter.write(self.__str__())
        xmlwriter.close_tag(newline=False)


class LineMixingRecord:
    """Represents a LineMixingRecord object.

    See online ARTS documentation for object details.

    """

    def __init__(self, tag=None, quantumnumberrecord=None, data=None):

        self.tag = tag
        self.quantumnumberrecord = quantumnumberrecord
        self.data = LineMixing(data)

    @property
    def tag(self):
        """:class:`SpeciesTag`"""
        return self._tag

    @property
    def quantumnumberrecord(self):
        """:class:`QuantumNumberRecord`"""
        return self._quantumnumberrecord

    @property
    def data(self):
        """Lineshape parameters."""
        return self._data

    @tag.setter
    def tag(self, tag):
        if tag is None:
            self._tag = None
            return

        self._tag = SpeciesTag(tag)

    def __repr__(self):
        return self.tag + ' ' + str(self.quantumnumberrecord) + ' ' + \
            str(self.data)

    @quantumnumberrecord.setter
    def quantumnumberrecord(self, quantumnumberrecord):
        self._quantumnumberrecord = return_if_arts_type(
            quantumnumberrecord, 'QuantumNumberRecord')

    @data.setter
    def data(self, data):
        self._data = return_if_arts_type(data, 'LineMixing')

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a LineMixingRecord object from an existing file.
        """

        obj = cls()
        obj.tag = xmlelement[0].value()
        obj.quantumnumberrecord = xmlelement[1].value()
        obj.data = LineMixing(xmlelement[2].value())

        return obj

    def write_xml(self, xmlwriter, attr=None):
        """Write a LineMixingRecord object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("LineMixingRecord", attr)
        xmlwriter.write_xml(self.tag)
        xmlwriter.write_xml(self.quantumnumberrecord)
        xmlwriter.write_xml(self.data.data)
        xmlwriter.close_tag()


class PropagationMatrix:
    """Represents a PropagationMatrix object.

    See online ARTS documentation for object details.

    """

    def __init__(self, data=np.zeros((1))):
        n = len(data.shape)
        if n == 4:
            self.aa = data.shape[0]
            self.za = data.shape[1]
            self.nf = data.shape[2]
        elif n == 3:
            self.aa = 1
            self.za = data.shape[0]
            self.nf = data.shape[1]
        elif n == 2:
            self.aa = 1
            self.za = 1
            self.nf = data.shape[0]
        elif n != 1:
            raise RuntimeError("Bad input")
        else:
            self.aa = 1
            self.za = 1
            self.nf = 1

        if data.shape[-1] == 7:
            self.stokes = 4
        elif data.shape[-1] == 4:
            self.stokes = 3
        elif data.shape[-1] == 1 or data.shape[-1] == 2:
            self.stokes = data.shape[-1]
        else:
            raise RuntimeError("Bad input")

        self.data = data.reshape(self.aa, self.za, self.nf, data.shape[-1])
    
    def Kjj(self, aa=0, za=0):
        return self.data[aa, za, :, 0]
    
    def K12(self, aa=0, za=0):
        return self.data[aa, za, :, 1]
    
    def K13(self, aa=0, za=0):
        return self.data[aa, za, :, 2]
    
    def K14(self, aa=0, za=0):
        return self.data[aa, za, :, 3]
    
    def K23(self, aa=0, za=0):
        return self.data[aa, za, :, 4]
    
    def K24(self, aa=0, za=0):
        return self.data[aa, za, :, 5]
    
    def K34(self, aa=0, za=0):
        return self.data[aa, za, :, 6]
    
    def __add__(self, other):
        if isinstance(other, PropagationMatrix):
            return PropagationMatrix(self.data + other.data)
        else:
            return PropagationMatrix(self.data + other)
    __radd__ = __add__
    
    def __repr__(self):
        size = "Stokes Dim: {}, Freqs: {}, Zeniths: {}, Azimuths: {}".format(self.stokes, self.nf, self.za, self.aa)
        return "PropagationMatrix of size <{}>".format(size)
    __str__=__repr__

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a PropagationMatrix object from an existing file.
        """
        return cls(xmlelement[0].value())

    def write_xml(self, xmlwriter, attr=None):
        """Write a PropagationMatrix object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("PropagationMatrix", attr)
        xmlwriter.write_xml(self.data)
        xmlwriter.close_tag()


class StokesVector:
    """Represents a StokesVector object.

    See online ARTS documentation for object details.

    """

    def __init__(self, data=np.zeros((1))):
        n = len(data.shape)
        if n == 4:
            self.aa = data.shape[0]
            self.za = data.shape[1]
            self.nf = data.shape[2]
        elif n == 3:
            self.aa = 1
            self.za = data.shape[0]
            self.nf = data.shape[1]
        elif n == 2:
            self.aa = 1
            self.za = 1
            self.nf = data.shape[0]
        elif n != 1:
            raise RuntimeError("Bad input")
        else:
            self.aa = 1
            self.za = 1
            self.nf = 1

        if data.shape[-1] == 1 or data.shape[-1] == 2 or \
           data.shape[-1] == 3 or data.shape[-1] == 4:
            self.stokes = data.shape[-1]
        else:
            raise RuntimeError("Bad input")

        self.data = data.reshape(self.aa, self.za, self.nf, data.shape[-1])

    @classmethod
    def from_xml(cls, xmlelement):
        """Loads a StokesVector object from an existing file.
        """
        return cls(xmlelement[0].value())

    def write_xml(self, xmlwriter, attr=None):
        """Write a StokesVector object to an ARTS XML file.
        """
        if attr is None:
            attr = {}

        xmlwriter.open_tag("StokesVector", attr)
        xmlwriter.write_xml(self.data)
        xmlwriter.close_tag()


try:
    from .utils import return_if_arts_type
    from .utils import as_quantumnumbers
    from .internals import PartitionFunctions
    from .internals import ARTSCAT5
    from .internals import Rational
    from .internals import LineMixing
except:
    from typhon.arts.utils import return_if_arts_type, as_quantumnumbers
    from typhon.arts.internals import PartitionFunctions, ARTSCAT5, Rational, \
        LineMixing



# -*- encoding: utf-8 -*-
import os
from tempfile import mkstemp

import numpy as np
import pytest

from typhon.arts import griddedfield, xml


def _create_tensor(n):
    """Create a tensor of dimension n.

    Create a tensor with n dimensions with two entries in each dimension.
    The tensor is filled with increasing integers starting with 0.

    Args:
        n (int): number of dimensions

    Returns:
        np.ndarray: n-dimensional tensor

    """
    return np.arange(2 ** n).reshape(2 * np.ones(n).astype(int))


def _get_griddedfield_type(dim):
    """Return the appropriate `GriddedField` type for given dimension."""
    return {
        1: griddedfield.GriddedField1,
        2: griddedfield.GriddedField2,
        3: griddedfield.GriddedField3,
        4: griddedfield.GriddedField4,
        5: griddedfield.GriddedField5,
        6: griddedfield.GriddedField6,
    }.get(dim)


class TestGriddedFieldUsage:
    ref_dir = os.path.join(os.path.dirname(__file__), "reference", "")

    @pytest.mark.parametrize("dim", range(1, 7))
    def test_check_init(self, dim):
        """Test initialisation of GriddedFields."""
        cls = _get_griddedfield_type(dim)
        assert cls().dimension == dim

    def test_check_dimension1(self):
        """Test if grid and data dimension agree (positive)."""
        gf3 = griddedfield.GriddedField3()
        gf3.grids = [np.arange(5), np.arange(5), []]
        gf3.gridnames = ["A", "B", "C"]
        gf3.data = np.ones((5, 5, 1))
        assert gf3.check_dimension() is True

    def test_check_dimension2(self):
        """Test if grid and data dimension agree (negative)."""
        gf3 = griddedfield.GriddedField3()
        gf3.grids = [np.arange(5), np.arange(5), []]
        gf3.gridnames = ["A", "B", "C"]

        # It shold not be allowed to set a Matrix as data in a GriddedField3.
        with pytest.raises(TypeError):
            gf3.data = np.ones((5, 5))

    def test_data(self):
        """Test setting and getting of data. """
        reference = np.random.randn(10, 10, 10)
        gf3 = griddedfield.GriddedField3()
        gf3.data = reference
        assert np.array_equal(gf3.data, reference)

    def test_name_setter(self):
        """Test name setter and getter."""
        reference = 'TestName'
        gf = griddedfield.GriddedField1()
        gf.name = reference
        assert gf.name == reference

    def _setup_gf2(self):
        """Helper for test_to_dict and test_to_xarray"""
        gf2 = griddedfield.GriddedField2()
        gf2.grids = [np.ones(5), np.zeros(5)]
        gf2.gridnames = ["ones", "zeros"]
        gf2.data = np.ones((5, 5))
        gf2.name = "semprini"
        return gf2

    def test_to_dict(self):
        """Test the conversion into a dictionary."""
        gf2 = self._setup_gf2()
        d = gf2.to_dict()

        res = (np.array_equal(d['ones'], np.ones(5)) and
               np.array_equal(d['zeros'], np.zeros(5)) and
               np.array_equal(d['data'], np.ones((5, 5))))

        assert res is True

    def test_to_xarray(self):
        """Test the conversion into xarray DataArray"""
        gf2 = self._setup_gf2()

        da = gf2.to_xarray()

        assert (da.name == "semprini" and
                da.dims == ("ones", "zeros") and
                np.array_equal(da.coords["zeros"], np.zeros(5)) and
                np.array_equal(da.coords["ones"], np.ones(5)) and
                np.array_equal(da.values, np.ones((5, 5))))

    @pytest.mark.parametrize('nametype', [float(), int()])
    def test_name_type(self, nametype):
        """Test if only names of type str are accepted."""
        with pytest.raises(TypeError):
            gf = griddedfield.GriddedField1()
            gf.name = nametype

    def test_shape(self):
        """Test return of data shape."""
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')

        assert gf3.shape == gf3.data.shape == (2, 2, 2)

    def test_data_subscription(self):
        """Test direct data subscription."""
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')

        assert gf3[0, 1, 0] == gf3.data[0, 1, 0]

    def test_slicing(self):
        """Test GriddedField slicing."""
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')

        # Create new GriddedField which is a sliced subset of the initial one.
        gf_sliced = gf3.extract_slice(slice(1, None), axis=1)

        assert np.allclose(gf3.data[:, 1:, :], gf_sliced.data)

    def test_repr(self):
        """Test string represenation of GriddedField objects."""
        str(xml.load(self.ref_dir + 'GriddedField3.xml'))

    def test_repr_empty(self):
        """Test string represenation of empty GriddedField objects."""
        str(griddedfield.GriddedField1())

    def test_get(self):
        """Test the get method for named fields."""
        gf1 = griddedfield.GriddedField1(
            grids=[['foo', 'bar']],
            data=np.array([42, 13]),
        )

        assert gf1.get('foo') == np.array([42])

    def test_get_default(self):
        """Test the GriddedField.get() behavior for non-existing fieldnames."""
        gf1 = griddedfield.GriddedField1(
            grids=[['dummy']],
            data=np.array([0]),
        )

        # Return given default, if a name is not existing.
        assert gf1.get('nonexisting', 42) == 42

        # If no default is specified, return `None`.
        assert gf1.get('nonexisting') is None

    def test_get_keepdims(self):
        """Test the dimension handling of the GriddedField.get()."""
        gf1 = griddedfield.GriddedField1(
            grids=[['foo', 'bar']],
            data=np.array([42, 13]),
        )

        assert gf1.get('foo').shape == (1,)
        assert gf1.get('foo', keep_dims=False).shape == tuple()

    def test_get_nofieldnames(self):
        """Test behavior if first grids is not ArrayOfString."""
        gf1 = griddedfield.GriddedField1(
            grids=[[0]],
            data=np.array([0]),
        )

        with pytest.raises(TypeError):
            gf1.get(0)

    def test_scaling(self):
        """Test the scaling of data in named fields."""
        gf1 = griddedfield.GriddedField1(
            grids=[['first_field', 'second_field']],
            data=np.array([1., 1.]),
        )

        gf1.scale('second_field', 0.1)

        # Check if values if *only* values of the second fields are scaled.
        assert gf1.data[0] == np.array([1])
        assert gf1.data[1] == np.array([0.1])

    def test_add_offset(self):
        """Test adding of offset to data in named fields."""
        gf1 = griddedfield.GriddedField1(
            grids=[['first_field', 'second_field']],
            data=np.array([1., 1.]),
        )

        gf1.add('second_field', 0.5)

        # Check if values if *only* values of the second fields are scaled.
        assert gf1.data[0] == np.array([1.])
        assert gf1.data[1] == np.array([1.5])

    def test_integer_scaling(self):
        """Test the scaling of integer data in named fields."""
        gf1 = griddedfield.GriddedField1(
            grids=[['first_field', 'second_field']],
            data=np.array([1, 1]),
        )

        gf1.scale('second_field', 0.1)

        # Check if values if *only* values of the second fields are scaled.
        assert gf1.data[0] == np.array([1])
        assert gf1.data[1] == np.array([0.1])

    def test_set(self):
        """Test the set method for named fields."""
        gf1 = griddedfield.GriddedField1(
            grids=[['zero', 'one']],
            data=np.array([0, 0]),
        )

        gf1.set('one', 1)

        assert gf1.data[1] == np.array([1])


class TestGriddedFieldLoad:
    ref_dir = os.path.join(os.path.dirname(__file__), "reference", "")

    def test_load_data(self):
        """Load reference XML file for GriddedField3 and check the data."""
        reference = _create_tensor(3)
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')
        test_data = gf3.data
        assert np.array_equal(test_data, reference)

    def test_load_grids(self):
        """Load reference XML file for GriddedField3 and check the grids."""
        reference = [np.arange(2)] * 3
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')
        test_data = gf3.grids
        assert all(np.allclose(a, b) for a, b in zip(test_data, reference))

    def test_load_gridnames(self):
        """Load reference XML file for GriddedField3 and check gridnames."""
        reference = ['grid1', 'grid2', 'grid3']
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')
        test_data = gf3.gridnames
        assert np.array_equal(test_data, reference)

    def test_load_dimension(self):
        """Load reference XML file for GriddedField3 and run check."""
        gf3 = xml.load(self.ref_dir + 'GriddedField3.xml')
        assert gf3.check_dimension()

    def test_equality(self):
        """Check the equality of two GriddedField objects."""
        # Create two different objects with same content.
        a = xml.load(self.ref_dir + 'GriddedField3.xml')
        b = xml.load(self.ref_dir + 'GriddedField3.xml')

        assert a == b

    def test_equality_empty(self):
        """Check the equality of two empty GriddedField objects."""
        # Create two different objects with same content.
        a = griddedfield.GriddedField3()
        b = griddedfield.GriddedField3()

        assert a == b

    def test_nonequality(self):
        """Check the non-equality of two GriddedField objects."""
        # Create two different objects with same content.
        a = xml.load(self.ref_dir + 'GriddedField3.xml')
        b = xml.load(self.ref_dir + 'GriddedField3.xml')

        a.name = 'foo'
        b.name = 'bar'

        assert a != b

    def test_copy(self):
        """Test copying of GriddedFields."""
        a = xml.load(self.ref_dir + 'GriddedField3.xml')
        b = a.copy()

        # GriddedFields should be equal but not the same object.
        assert a is not b and a == b

    def test_deepcopy(self):
        """Test deepcopying of GriddedField attributes."""
        a = xml.load(self.ref_dir + 'GriddedField3.xml')
        b = a.copy()

        # Grids should not be the same object.
        assert a.grids is not b.grids

    def test_from_xarray(self):
        a = xml.load(self.ref_dir + 'GriddedField3.xml')
        a.dataname = 'Testdata'
        da = a.to_xarray()
        b = griddedfield.GriddedField3.from_xarray(da)
        assert a == b


class TestGriddedFieldWrite:
    def setup_method(self):
        """Create a temporary file."""
        fd, self.f = mkstemp()
        os.close(fd)

    def teardown_method(self):
        """Delete temporary file."""
        os.remove(self.f)

    @pytest.mark.parametrize("dim", range(1, 7))
    def test_write_load_griddedfield(self, dim):
        gf = _get_griddedfield_type(dim)()
        gf.grids = [np.arange(2)] * dim
        gf.data = _create_tensor(dim)
        xml.save(gf, self.f)

        test_data = xml.load(self.f)

        assert np.array_equal(gf.data, test_data.data)



"""Collection of functions concerning the Optimal Estimation Method (OEM)."""

from typhon.retrieval.oem.common import *  # noqa
from typhon.retrieval.oem.error import *  # noqa


__all__ = [s for s in dir() if not s.startswith('_')]



""" The methods submodule.

This module exposes all ARTS workspace methods represented by WorkspaceMethod object.

The methods are loaded dynamically when the module is imported, which ensures that they
up to date with the current ARTS build.

Attributes:

     workspace_methods(dict): Dictionary containing all ARTS workspace methods.

"""

import ast
import re
import ctypes as c
import numpy as np


from typhon.arts.workspace.api       import arts_api
from typhon.arts.workspace.variables import WorkspaceVariable, group_ids, group_names
from typhon.arts.workspace import variables, workspace
from typhon.arts.workspace.output import CoutCapture

class WorkspaceMethod:
    """
    The WorkspaceMethod class represents ARTS workspace methods. Each workspace method
    provided a call function that forwards the call to the function towards the ARTS C API.

    Attributes:
        m_ids([int]):       Indices of supergeneric overloads of this WSM
        name(str):          The name of the method as defined in methods.cc
        description(str):   The documentation of the method as defined in methods.cc
        outs([int]):        Indices of the output variables of the method.
        n_out(int):         The number of output arguments.
        n_g_out(int):       The number of generic outputs.
        g_out([str]):       The names of the generic output arguments.
        g_out_types([dict]): List of dicts associating the name of a generic output
                             with its types for each of the supergeneric overloads.
        n_in(int):          The number of input arguments.
        ins([int]):         The indices of the input arguments of the WSM.
        n_g_in(int):        The number of generic input arguments.
        g_in_types([dict]): List of dicts associating the name of a generic input to the
                            expected type for each supergeneric overload of the method.
        g_in_default(dict)  Dictionary containing the default values for each generic parameter.
        g_in([str]):        List of names of the generic input arguments.
    """

    # Regular expression to that matches <Group>Create WSMs
    create_regexp = re.compile(r"^(\w*)Create$")

    def __init__(self, m_id, name, description, outs, g_out_types, ins, g_in_types):
        """Create a WorkspaceMethod object from a given id, name, description and types of
        generic input and output arguments.

        Args:
        m_id(int):          The index identifying the method in the C API
        name(str):          The name of the method
        description(str):   Method documentation
        outs([int]):        List of indices of the (non-generic) output arguments of the method.
        g_out_types([int]): Group ids of the generic output types
        outs([int]):        List of indices of the (non-generic) input arguments of the method.
        g_in_types([int]):  Group ids of the generic input types
        """
        self.m_ids       = [m_id]
        self.name        = name
        self.description = description

        # Output
        self.outs  = outs
        self.n_out = len(outs)

        # Generic Output
        self.n_g_out     = len(g_out_types)
        self.g_out_types = [WorkspaceMethod.get_output_dict(m_id, g_out_types)]
        self.g_out       = [k for k in self.g_out_types[0]]

        # Input
        self.ins  = ins
        self.n_in = len(ins)

        # Generic Input
        self.n_g_in       = len(g_in_types)
        self.g_in_types   = [WorkspaceMethod.get_input_dict(m_id, g_in_types)]
        self.g_in_default = WorkspaceMethod.get_default_input_dict(m_id, g_in_types)
        self.g_in         = [k for k in self.g_in_types[0]]

        self.is_create = False
        if (WorkspaceMethod.create_regexp.match(name)):
                self.is_create = True

    def __repr__(self):
        return arts_api.method_print_doc(self.m_ids[0]).decode("utf8")

    def add_overload(self, m_ids, g_in_types, g_out_types):
        """ Add one or more overloads to a workspace method.

        Use this function to add a supergeneric overload to a WorkspaceMethod object
        so that is will be considered in overload resolution when call(...) is called.

        TODO: Simplify this to take a WorkspaceMethod object.

        Args:
            m_ids([int]): The method ids of the supergeneric ARTS WSMs which should be added
                          to the list of overloads.
            g_in_types ([dict]): List of dicts containing the mappings between argument
                                 names and indices of expected groups
            g_out_types ([dict]): List of dicts containing the mappings between argument
                                  names and indices of expected groups
        """
        self.m_ids += m_ids
        self.g_in_types += g_in_types
        self.g_out_types += g_out_types

    @staticmethod
    def get_input_dict(m_id, in_types):
        """Get mapping of names of generic input variables to indices of groups.

        Args:
            m_id(int):       Index of the method.
            in_types([int]): List of group indices of the generic input arguments.
        Return:
            dict: The mapping.
        """
        res = dict()
        for i,t in enumerate(in_types):
            res[arts_api.get_method_g_in(m_id, i).decode("utf8")] = t
        return res

    @staticmethod
    def get_default_input_dict(m_id, g_in):
        """Get dict mapping names of generic input arguments to default values.

        Is None if no default value for a given generic input is given.

        Args:
            m_id(int):   Index of the method.
            g_in([str]): Names of the generic input arguments.
        Return:
            dict: The mapping.
        """
        res = dict()
        for i,t in enumerate(g_in):
            k = arts_api.get_method_g_in(m_id, i).decode("utf8")
            d = arts_api.get_method_g_in_default(m_id, i).decode("utf8")
            if d == "":
                pass
            elif d[0] == "@":
                d = None
            elif d == "Inf":
                res[k] = np.float64("inf")
            else:
                try:
                    d = WorkspaceVariable.convert(group_names[t], ast.literal_eval(d))
                    res[k] = d
                except:
                    res[k] = d
        return res

    @staticmethod
    def get_output_dict(m_id, out_types):
        """Get mapping of names of generic output variables to indices of groups.

        Args:
            m_id(int):       Index of the method.
            in_types([int]): List of group indices of the generic output arguments.
        Return:
            dict: The mapping.
        """
        res = dict()
        for i,t in enumerate(out_types):
            res[arts_api.get_method_g_out(m_id, i).decode("utf8")] = t
        return res

    def _parse_output_input_lists(self, ws, args, kwargs):
        n_args = self.n_g_out + self.n_g_in

        ins  = self.ins[:]
        outs = self.outs[:]
        temps = []

        # Add positional arguments to kwargs
        if (len(args)) and (len(args)) < self.n_g_out + self.n_in:
            raise Exception("Only " + str(len(args)) + " positional arguments provided " +
                            "but WSM " + self.name + "requires at least "
                            + str(self.n_g_out + self.n_in))
        for j in range(len(args)):
            if j < self.n_g_out:
                name = self.g_out[j]
                try:
                    kwargs[name] = args[j]
                except:
                    raise Exception("Generic parameter " + str(name) + " set twice.")
            elif j < self.n_g_out + self.n_in:
                k = j - self.n_g_out
                if type(args[j]) == WorkspaceVariable:
                    ins[k] = args[j].ws_id
                else:
                    temps.append(ws.add_variable(args[j]))
                    ins[k] = temps[-1].ws_id
                if self.ins[k] in outs:
                    # Need to replace variable also in output if its used as both output
                    # and input.
                    outs[outs.index(self.ins[k])] = ins[k]

            elif j < self.n_g_out + self.n_in + self.n_g_in:
                name = self.g_in[j - self.n_g_out - self.n_in]
                try:
                    kwargs[name] = args[j]
                except:
                    raise Exception("Generic parameter " + str(name) + " set twice.")
            else:
                raise Exception(str(j) + " positional arguments given, but this WSM expects " +
                                str(j-1) + ".")

        # Check output argument names
        g_output_args = dict()
        for k in self.g_out:
            if not k in kwargs:
                raise Exception("WSM " + self.name + " needs generic output " + k)
            else:
                g_output_args[k] = kwargs[k]

        # Check input argument names
        g_input_args = dict()
        for k in self.g_in:
            if not k in kwargs:
                if k in self.g_in_default:
                    g_input_args[k] = self.g_in_default[k]
                else:
                    raise Exception("WSM " + self.name + " needs generic input " + k)
            else:
                g_input_args[k] = kwargs[k]

        # Resolve overload (if necessary).
        g_out_types = dict([(k,WorkspaceVariable.get_group_id(g_output_args[k]))
                            for k in self.g_out])
        g_in_types  = dict([(k,WorkspaceVariable.get_group_id(g_input_args[k]))
                            for k in self.g_in])
        m_id = self.m_ids[0]
        sg_index = 0

        if (len(self.m_ids) > 1):
            out_indices = [i for i,ts in enumerate(self.g_out_types) if ts == g_out_types]
            in_indices  = [i for i,ts in enumerate(self.g_in_types) if ts == g_in_types]
            sg_indices  = set(out_indices) & set(in_indices)

            if len(sg_indices) > 1:
                raise Exception("Could not uniquely resolve super-generic overload.")

            if len(sg_indices) == 0:
                raise Exception("Could not find super-generic overload matching"
                                + " the given groups.")

            sg_index = sg_indices.pop()
            m_id = self.m_ids[sg_index]

        # Combine input and output arguments into lists.
        arts_args_out = []
        for out in outs:
            arts_args_out.append(out)

        for name in self.g_out:
            arg = g_output_args[name]
            if not type(arg) == WorkspaceVariable:
                raise ValueError("Generic Output " + name + " must be an ARTS WSV.")
            group_id = arg.group_id
            expected = self.g_out_types[sg_index][name]
            if not group_id == expected:
                raise Exception("Generic output " + name + " expected to be of type "
                                + group_names[expected])
            arts_args_out.append(arg.ws_id)

        arts_args_in = []
        for i in ins:
            if not i in outs:
                arts_args_in.append(i)

        for name in self.g_in:
            arg = g_input_args[name]
            if type(arg) == WorkspaceVariable:
                arts_args_in.append(arg.ws_id)
            else:
                group_id = WorkspaceVariable.get_group_id(arg)
                expected = self.g_in_types[sg_index][name]
                if not group_id == expected:
                    raise Exception("Generic input " + name + " expected to be of type "
                                    + group_names[expected])
                temps.append(ws.add_variable(arg))
                arts_args_in.append(temps[-1].ws_id)
        return (m_id, arts_args_out, arts_args_in, temps)


    def create(self, ws, name = None):
        """
        Call to <Group>Create WSMs are handled differently. This method simply
        determines the group type from the function name and then add a variable of
        this type to the workspace ws. A handle of this variable is then added to
        as attribute to the typhon.arts.workspace.variables module.

        Args:
            ws(Workspace): Workspace object to add the variable to
            name(str):     Name of the variable to add to the workspace
        """
        group = WorkspaceMethod.create_regexp.match(self.name).group(1)
        group_id = group_ids[group]

        if not name:
            name = "__anonymous_" + str(len(ws._vars))
            ws_id = arts_api.add_variable(ws.ptr, group_id, name.encode())
        else:
            # Is there a WSM with that name?
            if name in workspace_methods.keys():
                raise Exception("A WSM with the name " + name + " already exists.")

            # Is there a WSV with that name?
            ws_id = arts_api.lookup_workspace_variable(name.encode())
            # Is yes, check that it is of the same group?
            if not ws_id == -1:
                v = arts_api.get_variable(ws_id)
                if not v.group == group_id:
                    raise Exception("A WSV with the name " + name + " but of goup "
                                    + group_names[v.group] + " already exists.")
            # Otherwise we add the variable.
            else:
                ws_id = arts_api.add_variable(ws.ptr, group_id, name.encode())

        wsv = WorkspaceVariable(ws_id, name, group, "User defined variable.", ws)
        setattr(variables, name, wsv)
        ws._vars[name] = wsv
        return wsv

    def call(*args, **kwargs):
        """ Execute workspace method.

        This method will execute the workspace method (args[0]) on the workspace object (args[1])
        interpreting the remaining arguments in `*args` and `**kwargs` as arguments.

        Positional arguments in `*args` are interpreted in order with output arguments coming
        first.

        Keyword arguments in kwargs are interpreted according to the name of the generic
        parameters of the ARTS WSM.

        Args:
        args(list): Positional arguments with the first argument being the WorkspaceMethod
        instance, i.e. self = args[0], the second the Workspace object (args[1]). The
        remaining arguments are interpreted as generic arguments to the ARTS WSM.
        kargs(dict): Keyword args are interpreted as named generic arguments to the ARTS WSM
        according to its definition in methods.cc.
        """

        self = args[0]

        if self.is_create:
            return self.create(*args[1:])

        ws   = args[1]

        (m_id, arts_args_out, arts_args_in, temps) = self._parse_output_input_lists(ws,
                                                                                    args[2:],
                                                                                    kwargs)

        # Execute WSM and check for errors.
        arg_out_ptr = c.cast((c.c_long * len(arts_args_out))(*arts_args_out), c.POINTER(c.c_long))
        arg_in_ptr = c.cast((c.c_long * len(arts_args_in))(*arts_args_in), c.POINTER(c.c_long))

        with CoutCapture(ws):
            e_ptr = arts_api.execute_workspace_method(ws.ptr, m_id,
                                                      len(arts_args_out),
                                                      arg_out_ptr,
                                                      len(arts_args_in),
                                                      arg_in_ptr)
        if (e_ptr):
            raise Exception("Call to ARTS WSM " + self.name + " failed with error: "
                           + e_ptr.decode("utf8").format())

        # Remove temporaries from workspace (in reverse order).
        for t in temps[::-1]:
            t.erase()


    def describe(self):
        """ Print WSM documentation. """
        print(self.description.format())

def iter_raw():
    """ Iterator returning a WorkspaceMethod object for each available ARTS WSM.

    This iterator returns super-generically overloaded methods several times.

    Yields:
        WorkspaceMethod: The next ARTS Workspace method as defined in methods.cc in
                         increasing order.
    """
    for i in range(arts_api.get_number_of_methods()):
        m            = arts_api.get_method(i)
        name         = m.name.decode("utf8")
        description  = m.description.decode("utf8")
        outs         = [m.outs[i] for i in range(m.n_out)]
        g_out_types  = [m.g_out_types[i] for i in range(m.n_g_out)]
        ins          = [m.ins[i] for i in range(m.n_in)]
        g_in_types   = [m.g_in_types[i] for i in range(m.n_g_in)]
        yield WorkspaceMethod(m.id, name, description, outs, g_out_types, ins, g_in_types)

def iter():
    """ Iterator returning a WorkspaceMethod object for each available ARTS WSM.

    This iterator returns overloaded Workspace methods, i.e. super-generically overloaded
    WSM are not returned multiple times.

    Yields:
        WorkspaceMethod: The next ARTS Workspace method as defined in methods.cc in
                         increasing order.
    """
    for k,m in workspace_methods:
        yield m

workspace_methods = dict()
for m in iter_raw():
    if m.name in workspace_methods:
        workspace_methods[m.name].add_overload(m.m_ids, m.g_in_types, m.g_out_types)
    else:
        workspace_methods[m.name] = m
